from __future__ import annotations

from datetime import timedelta
from typing import TYPE_CHECKING, Any

import ujson
from infrahub_sdk.protocols import CoreNode  # noqa: TC002
from prefect import flow
from prefect.automations import AutomationCore
from prefect.client.orchestration import get_client
from prefect.client.schemas.filters import DeploymentFilter, DeploymentFilterName
from prefect.events.actions import (
    RunDeployment,
)
from prefect.events.schemas.automations import EventTrigger, Posture
from prefect.events.schemas.events import ResourceSpecification
from prefect.logging import get_run_logger

from infrahub.core.constants import ComputedAttributeKind, InfrahubKind
from infrahub.core.registry import registry
from infrahub.git.repository import get_initialized_repo
from infrahub.services import services
from infrahub.support.macro import MacroDefinition
from infrahub.workflows.catalogue import (
    PROCESS_COMPUTED_MACRO,
    QUERY_COMPUTED_ATTRIBUTE_TRANSFORM_TARGETS,
    TRIGGER_UPDATE_JINJA_COMPUTED_ATTRIBUTES,
    TRIGGER_UPDATE_PYTHON_COMPUTED_ATTRIBUTES,
    UPDATE_COMPUTED_ATTRIBUTE_TRANSFORM,
)
from infrahub.workflows.utils import add_tags, wait_for_schema_to_converge

from .constants import (
    PROCESS_AUTOMATION_NAME,
    PROCESS_JINJA2_AUTOMATION_NAME_PREFIX,
    PROCESS_PYTHON_AUTOMATION_NAME_PREFIX,
    QUERY_AUTOMATION_NAME,
    QUERY_AUTOMATION_NAME_PREFIX,
)
from .models import ComputedAttributeAutomations, PythonTransformComputedAttribute, PythonTransformTarget

if TYPE_CHECKING:
    import logging

    from infrahub.core.schema.computed_attribute import ComputedAttribute
    from infrahub.services import InfrahubServices

UPDATE_ATTRIBUTE = """
mutation UpdateAttribute(
    $id: String!,
    $kind: String!,
    $attribute: String!,
    $value: String!
  ) {
  InfrahubUpdateComputedAttribute(
    data: {id: $id, attribute: $attribute, value: $value, kind: $kind}
  ) {
    ok
  }
}
"""


@flow(
    name="process_computed_attribute_transform",
    flow_run_name="Process computed attribute for {computed_attribute_kind}.{computed_attribute_name}",
)
async def process_transform(
    branch_name: str,
    node_kind: str,
    object_id: str,
    computed_attribute_name: str,  # pylint: disable=unused-argument
    computed_attribute_kind: str,  # pylint: disable=unused-argument
    updated_fields: list[str] | None = None,  # pylint: disable=unused-argument
) -> None:
    await add_tags(branches=[branch_name], nodes=[object_id])

    service = services.service
    schema_branch = registry.schema.get_schema_branch(name=branch_name)
    node_schema = schema_branch.get_node(name=node_kind, duplicate=False)
    transform_attributes: dict[str, ComputedAttribute] = {}
    for attribute in node_schema.attributes:
        if attribute.computed_attribute and attribute.computed_attribute.kind == ComputedAttributeKind.TRANSFORM_PYTHON:
            transform_attributes[attribute.name] = attribute.computed_attribute

    if not transform_attributes:
        return

    for attribute_name, transform_attribute in transform_attributes.items():
        transform = await service.client.get(
            kind="CoreTransformPython",
            branch=branch_name,
            id=transform_attribute.transform,
            prefetch_relationships=True,
            populate_store=True,
        )
        if not transform:
            continue

        repo_node = await service.client.get(
            kind=transform.repository.peer.typename,
            branch=branch_name,
            id=transform.repository.peer.id,
            raise_when_missing=True,
        )

        repo = await get_initialized_repo(
            repository_id=transform.repository.peer.id,
            name=transform.repository.peer.name.value,
            service=service,
            repository_kind=transform.repository.peer.typename,
            commit=repo_node.commit.value,
        )

        data = await service.client.query_gql_query(
            name=transform.query.peer.name.value,
            branch_name=branch_name,
            variables={"id": object_id},
            update_group=True,
            subscribers=[object_id],
        )

        transformed_data = await repo.execute_python_transform(
            branch_name=branch_name,
            commit=repo_node.commit.value,
            location=f"{transform.file_path.value}::{transform.class_name.value}",
            data=data,
            client=service.client,
        )

        await service.client.execute_graphql(
            query=UPDATE_ATTRIBUTE,
            variables={
                "id": object_id,
                "kind": node_kind,
                "attribute": attribute_name,
                "value": transformed_data,
            },
            branch_name=branch_name,
        )


@flow(
    name="trigger_update_python_computed_attributes",
    flow_run_name="Trigger updates for computed attributes on branch {branch_name} for {computed_attribute_kind}.{computed_attribute_name}",
)
async def trigger_update_python_computed_attributes(
    branch_name: str,
    computed_attribute_name: str,
    computed_attribute_kind: str,
) -> None:
    service = services.service
    await add_tags(branches=[branch_name])

    nodes = await service.client.all(kind=computed_attribute_kind, branch=branch_name)

    for node in nodes:
        await service.workflow.submit_workflow(
            workflow=UPDATE_COMPUTED_ATTRIBUTE_TRANSFORM,
            parameters={
                "branch_name": branch_name,
                "node_kind": computed_attribute_kind,
                "object_id": node.id,
                "computed_attribute_name": computed_attribute_name,
                "computed_attribute_kind": computed_attribute_kind,
            },
        )


@flow(
    name="process_computed_attribute_value_jinja2",
    flow_run_name="Update value for computed attribute {attribute_name}",
)
async def update_computed_attribute_value_jinja2(
    branch_name: str, obj: CoreNode, attribute_name: str, template_value: str
) -> None:
    log = get_run_logger()
    service = services.service

    await add_tags(branches=[branch_name], nodes=[obj.id], db_change=True)

    macro_definition = MacroDefinition(macro=template_value)
    my_filter = {}
    for variable in macro_definition.variables:
        components = variable.split("__")
        if len(components) == 2:
            property_name = components[0]
            property_value = components[1]
            attribute_property = getattr(obj, property_name)
            my_filter[variable] = getattr(attribute_property, property_value)
        elif len(components) == 3:
            relationship_name = components[0]
            property_name = components[1]
            property_value = components[2]
            relationship = getattr(obj, relationship_name)
            try:
                attribute_property = getattr(relationship.peer, property_name)
                my_filter[variable] = getattr(attribute_property, property_value)
            except ValueError:
                my_filter[variable] = ""

    value = macro_definition.render(variables=my_filter)
    existing_value = getattr(obj, attribute_name).value
    if value == existing_value:
        log.debug(f"Ignoring to update {obj} with existing value on {attribute_name}={value}")
        return

    await service.client.execute_graphql(
        query=UPDATE_ATTRIBUTE,
        variables={
            "id": obj.id,
            "kind": obj.get_kind(),
            "attribute": attribute_name,
            "value": value,
        },
        branch_name=branch_name,
    )
    log.info(f"Updating computed attribute {obj.get_kind()}.{attribute_name}='{value}' ({obj.id})")


@flow(
    name="process_computed_attribute_jinja2",
    flow_run_name="Process computed attribute for {computed_attribute_kind}.{computed_attribute_name}",
)
async def process_jinja2(
    branch_name: str,
    node_kind: str,
    object_id: str,
    computed_attribute_name: str,
    computed_attribute_kind: str,
    updated_fields: str | None = None,
) -> None:
    log = get_run_logger()
    service = services.service

    await add_tags(branches=[branch_name])
    updates: list[str] = []
    if isinstance(updated_fields, str):
        updates = ujson.loads(updated_fields)

    target_branch_schema = (
        branch_name if branch_name in registry.get_altered_schema_branches() else registry.default_branch
    )
    schema_branch = registry.schema.get_schema_branch(name=target_branch_schema)
    await service.client.schema.all(branch=branch_name, refresh=True)

    computed_macros = [
        attrib
        for attrib in schema_branch.computed_attributes.get_impacted_jinja2_targets(kind=node_kind, updates=updates)
        if attrib.kind == computed_attribute_kind and attrib.attribute.name == computed_attribute_name
    ]
    for computed_macro in computed_macros:
        found: list[CoreNode] = []
        for id_filter in computed_macro.node_filters:
            filters = {id_filter: object_id}
            nodes: list[CoreNode] = await service.client.filters(
                kind=computed_macro.kind,
                branch=branch_name,
                prefetch_relationships=True,
                populate_store=True,
                **filters,
            )
            found.extend(nodes)

        if not found:
            log.debug("No nodes found that requires updates")

        template_string = "n/a"
        if computed_macro.attribute.computed_attribute and computed_macro.attribute.computed_attribute.jinja2_template:
            template_string = computed_macro.attribute.computed_attribute.jinja2_template

        batch = await service.client.create_batch()
        for node in found:
            batch.add(
                task=update_computed_attribute_value_jinja2,
                branch_name=branch_name,
                obj=node,
                attribute_name=computed_macro.attribute.name,
                template_value=template_string,
            )

        _ = [response async for _, response in batch.execute()]


@flow(
    name="trigger_update_jinja2_computed_attributes",
    flow_run_name="Trigger updates for computed attributes for {computed_attribute_kind}.{computed_attribute_name}",
)
async def trigger_update_jinja2_computed_attributes(
    branch_name: str,
    computed_attribute_name: str,
    computed_attribute_kind: str,
) -> None:
    service = services.service
    await add_tags(branches=[branch_name])

    nodes = await service.client.all(kind=computed_attribute_kind, branch=branch_name)

    for node in nodes:
        await service.workflow.submit_workflow(
            workflow=PROCESS_COMPUTED_MACRO,
            parameters={
                "branch_name": branch_name,
                "computed_attribute_name": computed_attribute_name,
                "computed_attribute_kind": computed_attribute_kind,
                "node_kind": computed_attribute_kind,
                "object_id": node.id,
            },
        )


@flow(name="computed-attribute-setup", flow_run_name="Setup computed attributes in task-manager")
async def computed_attribute_setup(branch_name: str | None = None) -> None:  # pylint: disable=too-many-statements
    service = services.service
    branch_name = branch_name or registry.default_branch

    await add_tags(branches=[branch_name])

    log = get_run_logger()
    await wait_for_schema_to_converge(branch_name=branch_name, service=service, log=log)

    branches_with_diff_from_main = registry.get_altered_schema_branches()
    schema_branch = registry.schema.get_schema_branch(name=branch_name)

    async with get_client(sync_client=False) as client:
        deployments = {
            item.name: item
            for item in await client.read_deployments(
                deployment_filter=DeploymentFilter(name=DeploymentFilterName(any_=[PROCESS_COMPUTED_MACRO.name]))
            )
        }
        if PROCESS_COMPUTED_MACRO.name not in deployments:
            raise ValueError("Unable to find the deployment for PROCESS_COMPUTED_MACRO")

        deployment_id_jinja = deployments[PROCESS_COMPUTED_MACRO.name].id

        automations = await client.read_automations()
        existing_computed_attr_automations = ComputedAttributeAutomations.from_prefect(
            automations=automations, prefix=PROCESS_JINJA2_AUTOMATION_NAME_PREFIX
        )
        automations_to_keep = []
        mapping = schema_branch.computed_attributes.get_jinja2_target_map()
        for computed_attribute, source_node_types in mapping.items():
            log.info(f"processing {computed_attribute.key_name}")
            scope = registry.default_branch

            match_criteria: dict[str, Any] = {"infrahub.node.kind": source_node_types}
            if branches_with_diff_from_main:
                match_criteria["infrahub.branch.name"] = [f"!{branch}" for branch in branches_with_diff_from_main]

            automation = AutomationCore(
                name=PROCESS_AUTOMATION_NAME.format(
                    prefix=PROCESS_JINJA2_AUTOMATION_NAME_PREFIX, identifier=computed_attribute.key_name, scope=scope
                ),
                description=f"Process value of the computed attribute for {computed_attribute.key_name} [{scope}] and branches with the same schema",
                enabled=True,
                trigger=EventTrigger(
                    posture=Posture.Reactive,
                    expect={"infrahub.node.*"},
                    within=timedelta(0),
                    match=ResourceSpecification(match_criteria),
                    threshold=1,
                ),
                actions=[
                    RunDeployment(
                        source="selected",
                        deployment_id=deployment_id_jinja,
                        parameters={
                            "branch_name": "{{ event.resource['infrahub.branch.name'] }}",
                            "node_kind": "{{ event.resource['infrahub.node.kind'] }}",
                            "object_id": "{{ event.resource['infrahub.node.id'] }}",
                            "computed_attribute_name": computed_attribute.attribute.name,
                            "computed_attribute_kind": computed_attribute.kind,
                            "updated_fields": "{{ event.payload['fields'] | tojson }}",
                        },
                        job_variables={},
                    )
                ],
            )

            if existing_computed_attr_automations.has(identifier=computed_attribute.key_name, scope=scope):
                existing = existing_computed_attr_automations.get(identifier=computed_attribute.key_name, scope=scope)
                await client.update_automation(automation_id=existing.id, automation=automation)
                automations_to_keep.append(existing.id)
                log.info(f"{computed_attribute.key_name} Updated")
            else:
                automation_id = await client.create_automation(automation=automation)
                automations_to_keep.append(automation_id)
                log.info(f"{computed_attribute.key_name} Created")

            if branch_name == registry.default_branch:
                await service.workflow.submit_workflow(
                    workflow=TRIGGER_UPDATE_JINJA_COMPUTED_ATTRIBUTES,
                    parameters={
                        "branch_name": registry.default_branch,
                        "computed_attribute_name": computed_attribute.attribute.name,
                        "computed_attribute_kind": computed_attribute.kind,
                    },
                )

        for diff_branch in branches_with_diff_from_main:
            schema_branch = registry.schema.get_schema_branch(name=diff_branch)

            mapping = schema_branch.computed_attributes.get_jinja2_target_map()
            for computed_attribute, source_node_types in mapping.items():
                log.info(f"processing {computed_attribute.key_name}")

                automation = AutomationCore(
                    name=PROCESS_AUTOMATION_NAME.format(
                        prefix=PROCESS_PYTHON_AUTOMATION_NAME_PREFIX,
                        identifier=computed_attribute.key_name,
                        scope=diff_branch,
                    ),
                    description=f"Process value of the computed attribute for {computed_attribute.key_name} [{diff_branch}]",
                    enabled=True,
                    trigger=EventTrigger(
                        posture=Posture.Reactive,
                        expect={"infrahub.node.*"},
                        within=timedelta(0),
                        match=ResourceSpecification(
                            {
                                "infrahub.node.kind": source_node_types,
                                "infrahub.branch.name": diff_branch,
                            }
                        ),
                        threshold=1,
                    ),
                    actions=[
                        RunDeployment(
                            source="selected",
                            deployment_id=deployment_id_jinja,
                            parameters={
                                "branch_name": "{{ event.resource['infrahub.branch.name'] }}",
                                "node_kind": "{{ event.resource['infrahub.node.kind'] }}",
                                "object_id": "{{ event.resource['infrahub.node.id'] }}",
                                "computed_attribute_name": computed_attribute.attribute.name,
                                "computed_attribute_kind": computed_attribute.kind,
                                "updated_fields": "{{ event.payload['fields'] | tojson }}",
                            },
                            job_variables={},
                        )
                    ],
                )

                if existing_computed_attr_automations.has(identifier=computed_attribute.key_name, scope=diff_branch):
                    existing = existing_computed_attr_automations.get(
                        identifier=computed_attribute.key_name, scope=diff_branch
                    )
                    await client.update_automation(automation_id=existing.id, automation=automation)
                    automations_to_keep.append(existing.id)
                    log.info(f"{computed_attribute.key_name} Updated")
                else:
                    automation_id = await client.create_automation(automation=automation)
                    automations_to_keep.append(automation_id)
                    log.info(f"{computed_attribute.key_name} Created")

                if branch_name == diff_branch:
                    await service.workflow.submit_workflow(
                        workflow=TRIGGER_UPDATE_JINJA_COMPUTED_ATTRIBUTES,
                        parameters={
                            "branch_name": branch_name,
                            "computed_attribute_name": computed_attribute.attribute.name,
                            "computed_attribute_kind": computed_attribute.kind,
                        },
                    )

        automations_to_remove = existing_computed_attr_automations.return_obsolete(keep=automations_to_keep)
        for automation_to_remove in automations_to_remove:
            await client.delete_automation(automation_id=automation_to_remove)


@flow(
    name="computed-attribute-setup-python",
    flow_run_name="Setup computed attributes for Python transforms in task-manager",
)
async def computed_attribute_setup_python(
    branch_name: str | None = None,
    commit: str | None = None,  # pylint: disable=unused-argument
    trigger_updates: bool = True,
) -> None:
    log = get_run_logger()
    service = services.service

    branch_name = branch_name or registry.default_branch

    await add_tags(branches=[branch_name])

    await wait_for_schema_to_converge(branch_name=branch_name, service=service, log=log)

    computed_attributes = await _gather_python_transform_attributes(branch_name=branch_name, service=service, log=log)

    async with get_client(sync_client=False) as client:
        deployments = {
            item.name: item
            for item in await client.read_deployments(
                deployment_filter=DeploymentFilter(
                    name=DeploymentFilterName(
                        any_=[UPDATE_COMPUTED_ATTRIBUTE_TRANSFORM.name, QUERY_COMPUTED_ATTRIBUTE_TRANSFORM_TARGETS.name]
                    )
                )
            )
        }
        if UPDATE_COMPUTED_ATTRIBUTE_TRANSFORM.name not in deployments:
            raise ValueError("Unable to find the deployment for UPDATE_COMPUTED_ATTRIBUTE_TRANSFORM")
        if QUERY_COMPUTED_ATTRIBUTE_TRANSFORM_TARGETS.name not in deployments:
            raise ValueError("Unable to find the deployment for QUERY_COMPUTED_ATTRIBUTE_TRANSFORM_TARGETS")

        deployment_id_python = deployments[UPDATE_COMPUTED_ATTRIBUTE_TRANSFORM.name].id
        deployment_id_query = deployments[QUERY_COMPUTED_ATTRIBUTE_TRANSFORM_TARGETS.name].id

        automations = await client.read_automations()
        existing_computed_attr_process_automations = ComputedAttributeAutomations.from_prefect(
            automations=automations, prefix=f"{PROCESS_PYTHON_AUTOMATION_NAME_PREFIX}::{branch_name}::"
        )
        existing_computed_attr_query_automations = ComputedAttributeAutomations.from_prefect(
            automations=automations, prefix=f"{QUERY_AUTOMATION_NAME_PREFIX}::{branch_name}::"
        )

        automations_to_keep = []
        for computed_attribute in computed_attributes:
            log.info(f"processing {computed_attribute.computed_attribute.key_name}")
            scope = branch_name

            automation = AutomationCore(
                name=PROCESS_AUTOMATION_NAME.format(
                    prefix=PROCESS_PYTHON_AUTOMATION_NAME_PREFIX,
                    identifier=computed_attribute.computed_attribute.key_name,
                    scope=scope,
                ),
                description=f"Process value of the computed attribute for {computed_attribute.computed_attribute.key_name} [{scope}]",
                enabled=True,
                trigger=EventTrigger(
                    posture=Posture.Reactive,
                    expect={"infrahub.node.*"},
                    within=timedelta(0),
                    match=ResourceSpecification(
                        {
                            "infrahub.node.kind": [computed_attribute.computed_attribute.kind],
                            "infrahub.branch.name": branch_name,
                        }
                    ),
                    threshold=1,
                ),
                actions=[
                    RunDeployment(
                        source="selected",
                        deployment_id=deployment_id_python,
                        parameters={
                            "branch_name": "{{ event.resource['infrahub.branch.name'] }}",
                            "node_kind": "{{ event.resource['infrahub.node.kind'] }}",
                            "object_id": "{{ event.resource['infrahub.node.id'] }}",
                            "computed_attribute_name": computed_attribute.computed_attribute.attribute.name,
                            "computed_attribute_kind": computed_attribute.computed_attribute.kind,
                        },
                        job_variables={},
                    )
                ],
            )

            if existing_computed_attr_process_automations.has(
                identifier=computed_attribute.computed_attribute.key_name, scope=scope
            ):
                existing = existing_computed_attr_process_automations.get(
                    identifier=computed_attribute.computed_attribute.key_name, scope=scope
                )
                await client.update_automation(automation_id=existing.id, automation=automation)
                log.info(f"Process {computed_attribute.computed_attribute.key_name} Updated")
                automations_to_keep.append(existing.id)
            else:
                automation_id = await client.create_automation(automation=automation)
                automations_to_keep.append(automation_id)
                log.info(f"Process {computed_attribute.computed_attribute.key_name} Created")

            automation = AutomationCore(
                name=QUERY_AUTOMATION_NAME.format(
                    prefix=QUERY_AUTOMATION_NAME_PREFIX,
                    identifier=computed_attribute.computed_attribute.key_name,
                    scope=scope,
                ),
                description=f"Query the computed attribute targets for {computed_attribute.computed_attribute.key_name} [{scope}]",
                enabled=True,
                trigger=EventTrigger(
                    posture=Posture.Reactive,
                    expect={"infrahub.node.*"},
                    within=timedelta(0),
                    match=ResourceSpecification(
                        {
                            "infrahub.node.kind": computed_attribute.query_models,
                            "infrahub.branch.name": branch_name,
                        }
                    ),
                    threshold=1,
                ),
                actions=[
                    RunDeployment(
                        source="selected",
                        deployment_id=deployment_id_query,
                        parameters={
                            "branch_name": "{{ event.resource['infrahub.branch.name'] }}",
                            "node_kind": "{{ event.resource['infrahub.node.kind'] }}",
                            "object_id": "{{ event.resource['infrahub.node.id'] }}",
                        },
                        job_variables={},
                    )
                ],
            )

            if existing_computed_attr_query_automations.has(
                identifier=computed_attribute.computed_attribute.key_name, scope=scope
            ):
                existing = existing_computed_attr_query_automations.get(
                    identifier=computed_attribute.computed_attribute.key_name, scope=scope
                )
                await client.update_automation(automation_id=existing.id, automation=automation)
                automations_to_keep.append(existing.id)
                log.info(f"Query {computed_attribute.computed_attribute.key_name} Updated")
            else:
                automation_id = await client.create_automation(automation=automation)
                automations_to_keep.append(automation_id)
                log.info(f"Query {computed_attribute.computed_attribute.key_name} Created")

            if trigger_updates:
                await service.workflow.submit_workflow(
                    workflow=TRIGGER_UPDATE_PYTHON_COMPUTED_ATTRIBUTES,
                    parameters={
                        "branch_name": branch_name,
                        "computed_attribute_name": computed_attribute.computed_attribute.attribute.name,
                        "computed_attribute_kind": computed_attribute.computed_attribute.kind,
                    },
                )

        automations_to_remove = existing_computed_attr_process_automations.return_obsolete(keep=automations_to_keep)
        for automation_to_remove in automations_to_remove:
            await client.delete_automation(automation_id=automation_to_remove)

        automations_to_remove = existing_computed_attr_query_automations.return_obsolete(keep=automations_to_keep)
        for automation_to_remove in automations_to_remove:
            await client.delete_automation(automation_id=automation_to_remove)


@flow(
    name="computed-attribute-remove-python",
    flow_run_name="Remove Python based computed attributes on branch={branch_name}",
)
async def computed_attribute_remove_python(
    branch_name: str,
) -> None:
    async with get_client(sync_client=False) as client:
        automations = await client.read_automations()
        existing_computed_attr_process_automations = ComputedAttributeAutomations.from_prefect(
            automations=automations, prefix=f"{PROCESS_PYTHON_AUTOMATION_NAME_PREFIX}::{branch_name}::"
        )
        existing_computed_attr_query_automations = ComputedAttributeAutomations.from_prefect(
            automations=automations, prefix=f"{QUERY_AUTOMATION_NAME_PREFIX}::{branch_name}::"
        )

        for automation_id in existing_computed_attr_process_automations.all_automation_ids:
            await client.delete_automation(automation_id=automation_id)

        for automation_id in existing_computed_attr_query_automations.all_automation_ids:
            await client.delete_automation(automation_id=automation_id)


@flow(
    name="query-computed-attribute-transform-targets",
    flow_run_name="Query for potential targets of computed attributes for {node_kind}",
)
async def query_transform_targets(
    branch_name: str,
    node_kind: str,  # pylint: disable=unused-argument
    object_id: str,
) -> None:
    await add_tags(branches=[branch_name])
    service = services.service
    schema_branch = registry.schema.get_schema_branch(name=branch_name)
    targets = await service.client.execute_graphql(
        query=GATHER_GRAPHQL_QUERY_SUBSCRIBERS, variables={"members": [object_id]}, branch_name=branch_name
    )

    subscribers: list[PythonTransformTarget] = []

    for group in targets[InfrahubKind.GRAPHQLQUERYGROUP]["edges"]:
        for subscriber in group["node"]["subscribers"]["edges"]:
            subscribers.append(
                PythonTransformTarget(object_id=subscriber["node"]["id"], kind=subscriber["node"]["__typename"])
            )

    nodes_with_computed_attributes = schema_branch.computed_attributes.get_python_attributes_per_node()
    for subscriber in subscribers:
        if subscriber.kind in nodes_with_computed_attributes:
            for computed_attribute in nodes_with_computed_attributes[subscriber.kind]:
                await service.workflow.submit_workflow(
                    workflow=UPDATE_COMPUTED_ATTRIBUTE_TRANSFORM,
                    parameters={
                        "branch_name": branch_name,
                        "node_kind": subscriber.kind,
                        "object_id": subscriber.object_id,
                        "computed_attribute_name": computed_attribute.name,
                        "computed_attribute_kind": subscriber.kind,
                    },
                )


async def _gather_python_transform_attributes(
    branch_name: str, service: InfrahubServices, log: logging.Logger | logging.LoggerAdapter
) -> list[PythonTransformComputedAttribute]:
    schema_branch = registry.schema.get_schema_branch(name=branch_name)
    branches_with_diff_from_main = registry.get_altered_schema_branches()

    transform_attributes = schema_branch.computed_attributes.python_attributes_by_transform

    transform_names = list(transform_attributes.keys())
    if not transform_names:
        return []

    transforms = await service.client.filters(
        kind="CoreTransformPython",
        branch=branch_name,
        prefetch_relationships=True,
        populate_store=True,
        name__values=transform_names,
    )

    found_transforms_names = [transform.name.value for transform in transforms]
    for transform_name in transform_names:
        if transform_name not in found_transforms_names:
            log.warning(
                msg=f"The transform {transform_name} is assigned to a computed attribute but the transform could not be found in the database."
            )

    repositories = await service.client.get_list_repositories()
    computed_attributes: list[PythonTransformComputedAttribute] = []
    for transform in transforms:
        for attribute in transform_attributes[transform.name.value]:
            python_transform_computed_attribute = PythonTransformComputedAttribute(
                name=transform.name.value,
                repository_id=transform.repository.peer.id,
                repository_name=transform.repository.peer.name.value,
                repository_kind=transform.repository.peer.typename,
                query_name=transform.query.peer.name.value,
                query_models=transform.query.peer.models.value,
                computed_attribute=attribute,
                default_schema=branch_name not in branches_with_diff_from_main,
            )
            python_transform_computed_attribute.populate_branch_commit(
                repository_data=repositories.get(transform.repository.peer.name.value)
            )
            computed_attributes.append(python_transform_computed_attribute)

    return computed_attributes


GATHER_GRAPHQL_QUERY_SUBSCRIBERS = """
query GatherGraphQLQuerySubscribers($members: [ID!]) {
  CoreGraphQLQueryGroup(members__ids: $members) {
    edges {
      node {
        subscribers {
          edges {
            node {
              id
              __typename
            }
          }
        }
      }
    }
  }
}
"""



import pytest

from infrahub_sdk import InfrahubClient
from infrahub_sdk.exceptions import GraphQLError
from infrahub_sdk.node import InfrahubNode
from infrahub_sdk.schema.main import (
    AttributeKind,
    GenericSchema,
    NodeSchema,
    RelationshipDirection,
    RelationshipKind,
    SchemaRoot,
)
from infrahub_sdk.schema.main import AttributeSchema as Attr
from infrahub_sdk.schema.main import RelationshipSchema as Rel

NAMESPACE = "Testing"

TESTING_ANIMAL = f"{NAMESPACE}Animal"
TESTING_CAT = f"{NAMESPACE}Cat"
TESTING_DOG = f"{NAMESPACE}Dog"
TESTING_PERSON = f"{NAMESPACE}Person"


class SchemaAnimal:
    @pytest.fixture(scope="class")
    def schema_animal(self) -> GenericSchema:
        return GenericSchema(
            name="Animal",
            namespace=NAMESPACE,
            include_in_menu=True,
            human_friendly_id=["owner__name__value", "name__value"],
            uniqueness_constraints=[
                ["owner", "name__value"],
            ],
            attributes=[
                Attr(name="name", kind=AttributeKind.TEXT),
                Attr(name="weight", kind=AttributeKind.NUMBER, optional=True),
            ],
            relationships=[
                Rel(
                    name="owner",
                    kind=RelationshipKind.GENERIC,
                    optional=False,
                    peer=TESTING_PERSON,
                    cardinality="one",
                    identifier="person__animal",
                    direction=RelationshipDirection.OUTBOUND,
                ),
                Rel(
                    name="best_friend",
                    kind=RelationshipKind.GENERIC,
                    optional=True,
                    peer=TESTING_PERSON,
                    cardinality="one",
                    identifier="person__animal_friend",
                    direction=RelationshipDirection.OUTBOUND,
                ),
            ],
        )

    @pytest.fixture(scope="class")
    def schema_dog(self) -> NodeSchema:
        return NodeSchema(
            name="Dog",
            namespace=NAMESPACE,
            include_in_menu=True,
            inherit_from=[TESTING_ANIMAL],
            display_labels=["name__value", "breed__value"],
            attributes=[
                Attr(name="breed", kind=AttributeKind.TEXT, optional=False),
                Attr(name="color", kind=AttributeKind.COLOR, default_value="#444444", optional=True),
            ],
        )

    @pytest.fixture(scope="class")
    def schema_cat(self) -> NodeSchema:
        return NodeSchema(
            name="Cat",
            namespace=NAMESPACE,
            include_in_menu=True,
            inherit_from=[TESTING_ANIMAL],
            display_labels=["name__value", "breed__value", "color__value"],
            order_by=["name__value"],
            attributes=[
                Attr(name="breed", kind=AttributeKind.TEXT, optional=False),
                Attr(name="color", kind=AttributeKind.COLOR, default_value="#555555", optional=True),
            ],
        )

    @pytest.fixture(scope="class")
    def schema_person(self) -> NodeSchema:
        return NodeSchema(
            name="Person",
            namespace=NAMESPACE,
            include_in_menu=True,
            display_labels=["name__value"],
            default_filter="name__value",
            human_friendly_id=["name__value"],
            attributes=[
                Attr(name="name", kind=AttributeKind.TEXT, unique=True),
                Attr(name="height", kind=AttributeKind.NUMBER, optional=True),
            ],
            relationships=[
                Rel(
                    name="animals",
                    peer=TESTING_ANIMAL,
                    identifier="person__animal",
                    cardinality="many",
                    direction=RelationshipDirection.INBOUND,
                ),
                Rel(
                    name="best_friends",
                    peer=TESTING_ANIMAL,
                    identifier="person__animal_friend",
                    cardinality="many",
                    direction=RelationshipDirection.INBOUND,
                ),
            ],
        )

    @pytest.fixture(scope="class")
    def schema_base(
        self,
        schema_animal: GenericSchema,
        schema_person: NodeSchema,
        schema_cat: NodeSchema,
        schema_dog: NodeSchema,
    ) -> SchemaRoot:
        return SchemaRoot(version="1.0", generics=[schema_animal], nodes=[schema_person, schema_cat, schema_dog])

    @pytest.fixture(scope="class")
    async def load_schema(self, client: InfrahubClient, schema_base: SchemaRoot) -> None:
        resp = await client.schema.load(schemas=[schema_base.to_schema_dict()], wait_until_converged=True)
        if resp.errors:
            raise GraphQLError(errors=[resp.errors])

    @pytest.fixture(scope="class")
    async def person_liam(self, client: InfrahubClient) -> InfrahubNode:
        obj = await client.create(kind=TESTING_PERSON, name="Liam Walker", height=178)
        await obj.save()
        return obj

    @pytest.fixture(scope="class")
    async def person_sophia(self, client: InfrahubClient) -> InfrahubNode:
        obj = await client.create(kind=TESTING_PERSON, name="Sophia Walker", height=168)
        await obj.save()
        return obj

    @pytest.fixture(scope="class")
    async def person_ethan(self, client: InfrahubClient) -> InfrahubNode:
        obj = await client.create(kind=TESTING_PERSON, name="Ethan Carter", height=185)
        await obj.save()
        return obj

    @pytest.fixture(scope="class")
    async def cat_bella(self, client: InfrahubClient, person_ethan: InfrahubNode) -> InfrahubNode:
        obj = await client.create(kind=TESTING_CAT, name="Bella", breed="Bengal", color="#123456", owner=person_ethan)
        await obj.save()
        return obj

    @pytest.fixture(scope="class")
    async def cat_luna(self, client: InfrahubClient, person_ethan: InfrahubNode) -> InfrahubNode:
        obj = await client.create(kind=TESTING_CAT, name="Luna", breed="Siamese", color="#FFFFFF", owner=person_ethan)
        await obj.save()
        return obj

    @pytest.fixture(scope="class")
    async def dog_daisy(self, client: InfrahubClient, person_ethan: InfrahubNode) -> InfrahubNode:
        obj = await client.create(
            kind=TESTING_DOG, name="Daisy", breed="French Bulldog", color="#7B7D7D", owner=person_ethan
        )
        await obj.save()
        return obj

    @pytest.fixture(scope="class")
    async def dog_rocky(self, client: InfrahubClient, person_sophia: InfrahubNode) -> InfrahubNode:
        obj = await client.create(
            kind=TESTING_DOG, name="Rocky", breed="German Shepherd", color="#784212", owner=person_sophia
        )
        await obj.save()
        return obj



from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Optional, Union

from infrahub import lock
from infrahub.core.constants import GLOBAL_BRANCH_NAME
from infrahub.exceptions import BranchNotFoundError, DataTypeNotFoundError, InitializationError

if TYPE_CHECKING:
    from neo4j import AsyncSession

    from infrahub.core.attribute import BaseAttribute
    from infrahub.core.branch import Branch
    from infrahub.core.manager import NodeManager
    from infrahub.core.schema import MainSchemaTypes, NodeSchema
    from infrahub.core.schema.manager import SchemaManager
    from infrahub.database import InfrahubDatabase
    from infrahub.graphql.mutations.attribute import BaseAttributeCreate, BaseAttributeUpdate
    from infrahub.graphql.types import InfrahubObject
    from infrahub.permissions import PermissionBackend
    from infrahub.storage import InfrahubObjectStorage
    from infrahub.types import InfrahubDataType

# pylint: disable=too-many-public-methods


@dataclass
class Registry:
    id: Optional[str] = None
    attribute: dict[str, type[BaseAttribute]] = field(default_factory=dict)
    branch: dict[str, Branch] = field(default_factory=dict)
    node: dict = field(default_factory=dict)
    _default_branch: Optional[str] = None
    _default_ipnamespace: Optional[str] = None
    _schema: Optional[SchemaManager] = None
    default_graphql_type: dict[str, InfrahubObject | type[BaseAttribute]] = field(default_factory=dict)
    graphql_type: dict = field(default_factory=lambda: defaultdict(dict))
    data_type: dict[str, type[InfrahubDataType]] = field(default_factory=dict)
    input_type: dict[str, type[BaseAttributeCreate | BaseAttributeUpdate]] = field(default_factory=dict)
    attr_group: dict = field(default_factory=dict)
    _branch_object: Optional[type[Branch]] = None
    _manager: Optional[type[NodeManager]] = None
    _storage: Optional[InfrahubObjectStorage] = None
    permission_backends: list[PermissionBackend] = field(default_factory=list)

    @property
    def branch_object(self) -> type[Branch]:
        if not self._branch_object:
            raise InitializationError
        return self._branch_object

    @branch_object.setter
    def branch_object(self, value: type[Branch]) -> None:
        self._branch_object = value

    @property
    def default_branch(self) -> str:
        if not self._default_branch:
            raise InitializationError()

        return self._default_branch

    @default_branch.setter
    def default_branch(self, value: str) -> None:
        self._default_branch = value

    @property
    def default_ipnamespace(self) -> str:
        if not self._default_ipnamespace:
            raise InitializationError()

        return self._default_ipnamespace

    @default_ipnamespace.setter
    def default_ipnamespace(self, value: str) -> None:
        self._default_ipnamespace = value

    @property
    def schema(self) -> SchemaManager:
        if not self._schema:
            raise InitializationError()

        return self._schema

    @schema.setter
    def schema(self, value: SchemaManager) -> None:
        self._schema = value

    @property
    def manager(self) -> type[NodeManager]:
        if not self._manager:
            raise InitializationError

        return self._manager

    @manager.setter
    def manager(self, value: type[NodeManager]) -> None:
        self._manager = value

    @property
    def storage(self) -> InfrahubObjectStorage:
        if not self._storage:
            raise InitializationError

        return self._storage

    @storage.setter
    def storage(self, value: InfrahubObjectStorage) -> None:
        self._storage = value

    def schema_has_been_initialized(self) -> bool:
        if self._schema:
            return True
        return False

    def get_node_schema(self, name: str, branch: Optional[Union[Branch, str]] = None) -> NodeSchema:
        return self.schema.get_node_schema(name=name, branch=branch)

    def get_data_type(self, name: str) -> type[InfrahubDataType]:
        if name not in self.data_type:
            raise DataTypeNotFoundError(name=name)
        return self.data_type[name]

    def get_full_schema(
        self, branch: Optional[Union[Branch, str]] = None, duplicate: bool = True
    ) -> dict[str, MainSchemaTypes]:
        """Return all the nodes in the schema for a given branch."""
        return self.schema.get_full(branch=branch, duplicate=duplicate)

    def delete_all(self) -> None:
        self.branch = {}
        self.node = {}
        self._schema = None
        self._default_ipnamespace = None
        self.attr_group = {}
        self.data_type = {}
        self.attribute = {}
        self.input_type = {}

    def get_branch_from_registry(self, branch: Optional[Union[Branch, str]] = None) -> Branch:
        """Return a branch object from the registry based on its name.

        Args:
            branch (Optional[Union[Branch, str]]): Branch object or name of a branch

        Raises:
            BranchNotFoundError:

        Returns:
            Branch: A Branch Object
        """

        if branch and not isinstance(branch, str):
            return branch

        # if the name of the branch is not defined or not a string we used the default branch name
        if not branch or not isinstance(branch, str):
            branch = registry.default_branch

        # Try to get it from the registry
        #   if not present in the registry and if a session has been provided get it from the database directly
        #   and update the registry
        if branch in self.branch:
            return self.branch[branch]

        raise BranchNotFoundError(identifier=branch)

    async def get_branch(
        self,
        db: InfrahubDatabase,
        session: Optional[AsyncSession] = None,
        branch: Optional[Union[Branch, str]] = None,
    ) -> Branch:
        """Return a branch object based on its name.

        First the function will check in the registry
        if the Branch is not present, and if a session object has been provided
            it will attempt to retrieve the branch and its schema from the database.

        Args:
            branch (Optional[Union[Branch, str]]): Branch object or name of a branch
            session (Optional[AsyncSession], optional): AsyncSession to connect to the database. Defaults to None.

        Raises:
            BranchNotFoundError:

        Returns:
            Branch: A Branch Object
        """

        if branch and not isinstance(branch, str):
            return branch

        if not branch or not isinstance(branch, str):
            branch = registry.default_branch

        try:
            return self.get_branch_from_registry(branch=branch)
        except BranchNotFoundError:
            if not session and not db:
                raise

        async with lock.registry.local_schema_lock():
            obj = await self.branch_object.get_by_name(name=branch, db=db)
            registry.branch[branch] = obj

            # Pull the schema for this branch
            await registry.schema.load_schema(db=db, branch=obj)

        return obj

    def get_global_branch(self) -> Branch:
        return self.get_branch_from_registry(branch=GLOBAL_BRANCH_NAME)

    def get_altered_schema_branches(self) -> list[str]:
        """Return a list of branch names that has a different hash than the default branch"""
        default_branch = self.branch[registry.default_branch]
        return [
            name
            for name, branch in self.branch.items()
            if name not in [registry.default_branch, GLOBAL_BRANCH_NAME]
            and branch.active_schema_hash.main != default_branch.active_schema_hash.main
        ]


registry = Registry()



from __future__ import annotations

import copy
import hashlib
from collections import defaultdict
from itertools import chain
from typing import TYPE_CHECKING, Any, Callable, Iterator, Optional, Union

from infrahub_sdk.topological_sort import DependencyCycleExistsError, topological_sort
from infrahub_sdk.utils import compare_lists, deep_merge_dict, duplicates, intersection
from typing_extensions import Self

from infrahub.computed_attribute.constants import VALID_KINDS as VALID_COMPUTED_ATTRIBUTE_KINDS
from infrahub.core.constants import (
    RESERVED_ATTR_GEN_NAMES,
    RESERVED_ATTR_REL_NAMES,
    RESTRICTED_NAMESPACES,
    BranchSupportType,
    ComputedAttributeKind,
    HashableModelState,
    InfrahubKind,
    RelationshipCardinality,
    RelationshipDeleteBehavior,
    RelationshipDirection,
    RelationshipKind,
    SchemaElementPathType,
)
from infrahub.core.migrations import MIGRATION_MAP
from infrahub.core.models import (
    HashableModelDiff,
    SchemaBranchHash,
    SchemaDiff,
    SchemaUpdateValidationResult,
)
from infrahub.core.schema import (
    AttributePathParsingError,
    AttributeSchema,
    BaseNodeSchema,
    GenericSchema,
    MainSchemaTypes,
    NodeSchema,
    ProfileSchema,
    RelationshipSchema,
    SchemaAttributePath,
    SchemaRoot,
)
from infrahub.core.schema.definitions.core import core_profile_schema_definition
from infrahub.core.validators import CONSTRAINT_VALIDATOR_MAP
from infrahub.exceptions import SchemaNotFoundError, ValidationError
from infrahub.log import get_logger
from infrahub.support.macro import MacroDefinition
from infrahub.types import ATTRIBUTE_TYPES
from infrahub.utils import format_label
from infrahub.visuals import select_color

from .constants import INTERNAL_SCHEMA_NODE_KINDS, SchemaNamespace
from .schema_branch_computed import ComputedAttributes

log = get_logger()

if TYPE_CHECKING:
    from pydantic import ValidationInfo


# pylint: disable=redefined-builtin,too-many-public-methods,too-many-lines


class SchemaBranch:
    def __init__(
        self,
        cache: dict,
        name: str | None = None,
        data: dict[str, dict[str, str]] | None = None,
        computed_attributes: ComputedAttributes | None = None,
    ):
        self._cache: dict[str, Union[NodeSchema, GenericSchema]] = cache
        self.name: str | None = name
        self.nodes: dict[str, str] = {}
        self.generics: dict[str, str] = {}
        self.profiles: dict[str, str] = {}
        self.computed_attributes = computed_attributes or ComputedAttributes()

        if data:
            self.nodes = data.get("nodes", {})
            self.generics = data.get("generics", {})
            self.profiles = data.get("profiles", {})

    @classmethod
    def __get_validators__(cls) -> Iterator[Callable[..., Any]]:  # noqa: PLW3201
        yield cls.validate

    @classmethod
    def validate(cls, v: Any, info: ValidationInfo) -> Self:  # pylint: disable=unused-argument
        if isinstance(v, cls):
            return v
        if isinstance(v, dict):
            return cls.from_dict_schema_object(data=v)
        raise ValueError("must be a class or a dict")

    @property
    def node_names(self) -> list[str]:
        return list(self.nodes.keys())

    @property
    def generic_names(self) -> list[str]:
        return list(self.generics.keys())

    @property
    def profile_names(self) -> list[str]:
        return list(self.profiles.keys())

    def get_all_kind_id_map(self, exclude_profiles: bool = False) -> dict[str, str]:
        kind_id_map = {}
        if exclude_profiles:
            names = self.node_names + self.generic_names
        else:
            names = self.all_names
        for name in names:
            if name == InfrahubKind.NODE:
                continue
            item = self.get(name=name, duplicate=False)
            kind_id_map[name] = item.id
        return kind_id_map

    @property
    def all_names(self) -> list[str]:
        return self.node_names + self.generic_names + self.profile_names

    def get_hash(self) -> str:
        """Calculate the hash for this objects based on the content of nodes and generics.

        Since the object themselves are considered immutable we just need to use the hash from each object to calculate the global hash.
        """
        md5hash = hashlib.md5(usedforsecurity=False)
        for key, value in sorted(tuple(self.nodes.items()) + tuple(self.generics.items())):
            md5hash.update(str(key).encode())
            md5hash.update(str(value).encode())

        return md5hash.hexdigest()

    def get_hash_full(self) -> SchemaBranchHash:
        return SchemaBranchHash(main=self.get_hash(), nodes=self.nodes, generics=self.generics)

    def to_dict(self) -> dict[str, Any]:
        return {"nodes": self.nodes, "generics": self.generics, "profiles": self.profiles}

    def to_dict_schema_object(self, duplicate: bool = False) -> dict[str, dict[str, MainSchemaTypes]]:
        return {
            "nodes": {name: self.get(name, duplicate=duplicate) for name in self.nodes},
            "profiles": {name: self.get(name, duplicate=duplicate) for name in self.profiles},
            "generics": {name: self.get(name, duplicate=duplicate) for name in self.generics},
        }

    @classmethod
    def from_dict_schema_object(cls, data: dict) -> Self:
        type_mapping = {
            "nodes": NodeSchema,
            "generics": GenericSchema,
            "profiles": ProfileSchema,
        }

        cache: dict[str, MainSchemaTypes] = {}
        nodes: dict[str, dict[str, str]] = {"nodes": {}, "generics": {}, "profiles": {}}

        for node_type, node_class in type_mapping.items():
            for node_name, node_data in data[node_type].items():
                node: MainSchemaTypes = node_class(**node_data)
                node_hash = node.get_hash()
                nodes[node_type][node_name] = node_hash

                cache[node_hash] = node

        return cls(cache=cache, data=nodes)

    def diff(self, other: SchemaBranch) -> SchemaDiff:
        # Identify the nodes or generics that have been added or removed
        local_kind_id_map = self.get_all_kind_id_map(exclude_profiles=True)
        other_kind_id_map = other.get_all_kind_id_map(exclude_profiles=True)
        clean_local_ids = [id for id in local_kind_id_map.values() if id is not None]
        clean_other_ids = [id for id in other_kind_id_map.values() if id is not None]
        shared_ids = intersection(list1=clean_local_ids, list2=clean_other_ids)

        local_keys = [kind for kind, id in local_kind_id_map.items() if id not in shared_ids]
        other_keys = [kind for kind, id in other_kind_id_map.items() if id not in shared_ids]

        present_both, present_local_only, present_other_only = compare_lists(list1=local_keys, list2=other_keys)

        added_elements = {element: HashableModelDiff() for element in present_other_only}
        removed_elements = {element: HashableModelDiff() for element in present_local_only}
        schema_diff = SchemaDiff(added=added_elements, removed=removed_elements)

        # Process of the one that have been updated to identify the list of impacted fields
        for key in present_both:
            local_node = self.get(name=key, duplicate=False)
            other_node = other.get(name=key, duplicate=False)
            diff_node = other_node.diff(other=local_node)
            if diff_node.has_diff:
                schema_diff.changed[key] = diff_node

        reversed_map_local = dict(map(reversed, local_kind_id_map.items()))
        reversed_map_other = dict(map(reversed, other_kind_id_map.items()))

        for shared_id in shared_ids:
            local_node = self.get(name=reversed_map_local[shared_id], duplicate=False)
            other_node = other.get(name=reversed_map_other[shared_id], duplicate=False)
            diff_node = other_node.diff(other=local_node)
            if other_node.state == HashableModelState.ABSENT:
                schema_diff.removed[reversed_map_other[shared_id]] = None
            elif diff_node.has_diff:
                schema_diff.changed[reversed_map_other[shared_id]] = diff_node

        return schema_diff

    def update(self, schema: SchemaBranch) -> None:
        """Update another SchemaBranch into this one."""

        local_kinds = list(self.nodes.keys()) + list(self.generics.keys())
        other_kinds = list(schema.nodes.keys()) + list(schema.generics.keys())

        in_both, _, other_only = compare_lists(list1=local_kinds, list2=other_kinds)

        for item_kind in in_both:
            other_item = schema.get(name=item_kind)
            new_item = self.get(name=item_kind)
            new_item.update(other_item)
            self.set(name=item_kind, schema=new_item)

        for item_kind in other_only:
            other_item = schema.get(name=item_kind)
            self.set(name=item_kind, schema=other_item)

        # for item_kind in local_only:
        #     if item_kind in self.nodes:
        #         del self.nodes[item_kind]
        #     else:
        #         del self.generics[item_kind]

    def validate_node_deletions(self, diff: SchemaDiff) -> None:
        """Given a diff, check if a deleted node is still used in relationships of other nodes."""
        removed_schema_names = set(diff.removed.keys())
        for name in self.all_names:
            node = self.get(name=name, duplicate=False)
            for relationship in node.relationships:
                if relationship.peer in removed_schema_names:
                    raise ValueError(
                        f"'{relationship.peer}' has been removed but is still referenced in '{name}.{relationship.name}'; keep it or delete the "
                        "relationship"
                    )

    def validate_update(
        self, other: SchemaBranch, diff: SchemaDiff, enforce_update_support: bool = True
    ) -> SchemaUpdateValidationResult:
        result = SchemaUpdateValidationResult.init(
            diff=diff, schema=other, enforce_update_support=enforce_update_support
        )
        result.validate_all(migration_map=MIGRATION_MAP, validator_map=CONSTRAINT_VALIDATOR_MAP)
        return result

    def duplicate(self, name: Optional[str] = None) -> SchemaBranch:
        """Duplicate the current object but conserve the same cache."""
        return self.__class__(
            name=name,
            data=copy.deepcopy(self.to_dict()),
            cache=self._cache,
            computed_attributes=self.computed_attributes.duplicate(),
        )

    def set(self, name: str, schema: MainSchemaTypes) -> str:
        """Store a NodeSchema or GenericSchema associated with a specific name.

        The object will be stored in the internal cache based on its hash value.
        If a schema with the same name already exist, it will be replaced
        """
        schema_hash = schema.get_hash()
        if schema_hash not in self._cache:
            self._cache[schema_hash] = schema

        if "Node" in schema.__class__.__name__:
            self.nodes[name] = schema_hash
        elif "Generic" in schema.__class__.__name__:
            self.generics[name] = schema_hash
        elif "Profile" in schema.__class__.__name__:
            self.profiles[name] = schema_hash

        return schema_hash

    def get(self, name: str, duplicate: bool = True) -> MainSchemaTypes:
        """Access a specific NodeSchema or GenericSchema, defined by its kind.

        To ensure that no-one will ever change an object in the cache,
        by default the function always returns a copy of the object, not the object itself

        If duplicate is set to false, the real object will be returned.
        """
        key = None
        if name in self.nodes:
            key = self.nodes[name]
        elif name in self.generics:
            key = self.generics[name]
        elif name in self.profiles:
            key = self.profiles[name]

        if key and duplicate:
            return self._cache[key].duplicate()
        if key and not duplicate:
            return self._cache[key]

        raise SchemaNotFoundError(
            branch_name=self.name, identifier=name, message=f"Unable to find the schema {name!r} in the registry"
        )

    def get_node(self, name: str, duplicate: bool = True) -> NodeSchema:
        """Access a specific NodeSchema, defined by its kind."""
        item = self.get(name=name, duplicate=duplicate)
        if not isinstance(item, NodeSchema):
            raise ValueError(f"{name!r} is not of type NodeSchema")
        return item

    def get_generic(self, name: str, duplicate: bool = True) -> GenericSchema:
        """Access a specific GenericSchema, defined by its kind."""
        item = self.get(name=name, duplicate=duplicate)
        if not isinstance(item, GenericSchema):
            raise ValueError(f"{name!r} is not of type GenericSchema")
        return item

    def get_profile(self, name: str, duplicate: bool = True) -> ProfileSchema:
        """Access a specific ProfileSchema, defined by its kind."""
        item = self.get(name=name, duplicate=duplicate)
        if not isinstance(item, ProfileSchema):
            raise ValueError(f"{name!r} is not of type ProfileSchema")
        return item

    def delete(self, name: str) -> None:
        if name in self.nodes:
            del self.nodes[name]
        elif name in self.generics:
            del self.generics[name]
        elif name in self.profiles:
            del self.profiles[name]
        else:
            raise SchemaNotFoundError(
                branch_name=self.name, identifier=name, message=f"Unable to find the schema {name!r} in the registry"
            )

    def get_by_id(self, id: str, duplicate: bool = True) -> MainSchemaTypes:
        for name in self.all_names:
            node = self.get(name=name, duplicate=False)
            if node.id != id:
                continue
            if duplicate is False:
                return node
            return self.get(name=name, duplicate=True)

        raise SchemaNotFoundError(
            branch_name=self.name,
            identifier=id,
            message=f"Unable to find the schema with the id {id!r} in the registry",
        )

    def get_by_any_id(self, id: str) -> MainSchemaTypes:
        for name in self.all_names:
            node = self.get(name=name, duplicate=False)
            if node.id == id:
                return node

            # search in the attributes and the relationships
            try:
                node.get_attribute_by_id(id=id)
                return node

            except ValueError:
                pass

            try:
                node.get_relationship_by_id(id=id)
                return node
            except ValueError:
                pass

        raise SchemaNotFoundError(
            branch_name=self.name,
            identifier=id,
            message=f"Unable to find the schema with the id {id!r} or with an attribute or a relationship with this id",
        )

    def has(self, name: str) -> bool:
        try:
            self.get(name=name, duplicate=False)
            return True
        except SchemaNotFoundError:
            return False

    def get_all(self, include_internal: bool = False, duplicate: bool = True) -> dict[str, MainSchemaTypes]:
        """Retrieve everything in a single dictionary."""

        return {
            name: self.get(name=name, duplicate=duplicate)
            for name in self.all_names
            if include_internal or name not in INTERNAL_SCHEMA_NODE_KINDS
        }

    def get_namespaces(self, include_internal: bool = False) -> list[SchemaNamespace]:
        all_schemas = self.get_all(include_internal=include_internal, duplicate=False)
        namespaces: dict[str, SchemaNamespace] = {}
        for schema in all_schemas.values():
            if schema.namespace in namespaces:
                continue
            namespaces[schema.namespace] = SchemaNamespace(
                name=schema.namespace, user_editable=schema.namespace not in RESTRICTED_NAMESPACES
            )

        return list(namespaces.values())

    def get_schemas_for_namespaces(
        self, namespaces: Optional[list[str]] = None, include_internal: bool = False
    ) -> list[MainSchemaTypes]:
        """Retrive everything in a single dictionary."""
        all_schemas = self.get_all(include_internal=include_internal, duplicate=False)
        if namespaces:
            return [schema for schema in all_schemas.values() if schema.namespace in namespaces]
        return list(all_schemas.values())

    def get_schemas_by_rel_identifier(self, identifier: str) -> list[MainSchemaTypes]:
        nodes: list[RelationshipSchema] = []
        for node_name in list(self.nodes.keys()) + list(self.generics.keys()):
            node = self.get(name=node_name, duplicate=False)
            rel = node.get_relationship_by_identifier(id=identifier, raise_on_error=False)
            if rel:
                nodes.append(self.get(name=node_name, duplicate=True))
        return nodes

    def generate_fields_for_display_label(self, name: str) -> Optional[dict]:
        node = self.get(name=name, duplicate=False)
        if isinstance(node, (NodeSchema, ProfileSchema)):
            return node.generate_fields_for_display_label()

        fields: dict[str, Union[str, None, dict[str, None]]] = {}
        if isinstance(node, GenericSchema):
            for child_node_name in node.used_by:
                child_node = self.get(name=child_node_name, duplicate=False)
                resp = child_node.generate_fields_for_display_label()
                if not resp:
                    continue
                fields = deep_merge_dict(dicta=fields, dictb=resp)

        return fields or None

    def load_schema(self, schema: SchemaRoot) -> None:
        """Load a SchemaRoot object and store all NodeSchema or GenericSchema.

        In the current implementation, if a schema object present in the SchemaRoot already exist, it will be overwritten.
        """
        for item in schema.nodes + schema.generics:
            try:
                if item.id:
                    new_item = self.get_by_id(id=item.id)
                    if new_item.kind != item.kind:
                        self.delete(name=new_item.kind)
                else:
                    new_item = self.get(name=item.kind)

                if (new_item.is_node_schema and not item.is_node_schema) or (
                    new_item.is_generic_schema and not item.is_generic_schema
                ):
                    current_node_type = "Node" if new_item.is_node_schema else "Generic"
                    raise ValidationError(
                        f"{item.kind} already exist in the schema as a {current_node_type}. Either rename it or delete the existing one."
                    )
                new_item.update(item)
                self.set(name=item.kind, schema=new_item)
            except SchemaNotFoundError:
                self.set(name=item.kind, schema=item)

        for node_extension in schema.extensions.nodes:
            new_item = self.get(name=node_extension.kind)
            new_item.update(node_extension)
            self.set(name=node_extension.kind, schema=new_item)

    def process(self, validate_schema: bool = True) -> None:
        self.process_pre_validation()
        if validate_schema:
            self.process_validate()
        self.process_post_validation()

    def process_pre_validation(self) -> None:
        self.generate_identifiers()
        self.process_default_values()
        self.process_deprecations()
        self.process_cardinality_counts()
        self.process_inheritance()
        self.process_hierarchy()
        self.process_branch_support()
        self.manage_profile_schemas()
        self.manage_profile_relationships()

    def process_validate(self) -> None:
        self.validate_names()
        self.validate_kinds()
        self.validate_computed_attributes()
        self.validate_default_values()
        self.validate_count_against_cardinality()
        self.validate_identifiers()
        self.sync_uniqueness_constraints_and_unique_attributes()
        self.validate_uniqueness_constraints()
        self.validate_display_labels()
        self.validate_order_by()
        self.validate_default_filters()
        self.validate_parent_component()
        self.validate_human_friendly_id()
        self.validate_required_relationships()

    def process_post_validation(self) -> None:
        self.cleanup_inherited_elements()
        self.add_groups()
        self.add_hierarchy_generic()
        self.add_hierarchy_node()
        self.generate_weight()
        self.process_labels()
        self.process_dropdowns()
        self.process_relationships()
        self.process_human_friendly_id()

    def generate_identifiers(self) -> None:
        """Generate the identifier for all relationships if it's not already present."""
        for name in self.all_names:
            node = self.get(name=name, duplicate=False)
            rels_missing_identifier = [rel.name for rel in node.relationships if rel.identifier is None]
            if not rels_missing_identifier:
                continue

            node = node.duplicate()
            for rel in node.relationships:
                if rel.identifier:
                    continue
                rel.identifier = str("__".join(sorted([node.kind, rel.peer]))).lower()
            self.set(name=name, schema=node)

    def validate_identifiers(self) -> None:
        """Validate that all relationships have a unique identifier for a given model."""
        # Organize all the relationships per identifier and node
        rels_per_identifier: dict[str, dict[str, list[RelationshipSchema]]] = defaultdict(lambda: defaultdict(list))
        for name in self.all_names:
            node = self.get(name=name, duplicate=False)

            for rel in node.relationships:
                rels_per_identifier[rel.identifier][name].append(rel)

        valid_options = [
            [RelationshipDirection.BIDIR, RelationshipDirection.BIDIR],
            sorted([RelationshipDirection.INBOUND, RelationshipDirection.OUTBOUND]),
        ]

        for identifier, rels_per_kind in rels_per_identifier.items():
            # Per node kind, check if the directions are good
            for kind, rels in rels_per_kind.items():
                directions = sorted([rel.direction.value for rel in rels])
                if not (len(rels) == 1 or (len(rels) == 2 and directions == ["inbound", "outbound"])):
                    names_directions = [(rel.name, rel.direction.value) for rel in rels]
                    raise ValueError(
                        f"{kind}: Identifier of relationships must be unique for a given direction > {identifier!r} : {names_directions}"
                    ) from None

                # Continue if no other model is using this identifier
                if len(rels_per_kind) == 1:
                    continue

                # If this node has 2 relationships, BIDIRECTIONAL is not a valid option on the remote node
                if len(rels) == 2:
                    for rel in rels:
                        if (
                            rel.peer in list(rels_per_kind.keys())
                            and len(rels_per_kind[rel.peer]) == 1
                            and rels_per_kind[rel.peer][0].direction == RelationshipDirection.BIDIR
                        ):
                            raise ValueError(
                                f"{node.kind}: Incompatible direction detected on Reverse Relationship for {rel.name!r} ({identifier!r}) "
                                f" > {RelationshipDirection.BIDIR.value} "
                            ) from None

                elif (
                    len(rels) == 1
                    and rels[0].peer in list(rels_per_kind.keys())
                    and len(rels_per_kind[rels[0].peer]) == 1
                ):
                    peer_direction = rels_per_kind[rels[0].peer][0].direction
                    if sorted([peer_direction, rels[0].direction]) not in valid_options:
                        raise ValueError(
                            f"{node.kind}: Incompatible direction detected on Reverse Relationship for {rels[0].name!r} ({identifier!r})"
                            f" {rels[0].direction.value} <> {peer_direction.value}"
                        ) from None

    def validate_schema_path(
        self,
        node_schema: BaseNodeSchema,
        path: str,
        allowed_path_types: SchemaElementPathType,
        element_name: Optional[str] = None,
    ) -> SchemaAttributePath:
        error_header = f"{node_schema.kind}"
        error_header += f".{element_name}" if element_name else ""

        try:
            schema_attribute_path = node_schema.parse_schema_path(path=path, schema=self)
        except AttributePathParsingError as exc:
            raise ValueError(f"{error_header}: {exc}") from exc

        if not (SchemaElementPathType.ATTR & allowed_path_types) and schema_attribute_path.is_type_attribute:
            raise ValueError(f"{error_header}: this property only supports relationships not attributes")

        if not (SchemaElementPathType.REL & allowed_path_types) and schema_attribute_path.is_type_relationship:
            raise ValueError(f"{error_header}: this property only supports attributes, not relationships")

        if not (SchemaElementPathType.ATTR_NO_PROP & allowed_path_types) and schema_attribute_path.is_type_attribute:
            required_properties = tuple(
                schema_attribute_path.attribute_schema.get_class().get_allowed_property_in_path()
            )
            if schema_attribute_path.attribute_property_name not in required_properties:
                raise ValueError(
                    f"{error_header}: invalid attribute, it must end with one of the following properties:"
                    f" {', '.join(required_properties)}. (`{path}`)"
                )

        if schema_attribute_path.is_type_relationship:
            if schema_attribute_path.relationship_schema.cardinality == RelationshipCardinality.ONE:
                if not SchemaElementPathType.REL_ONE & allowed_path_types:
                    raise ValueError(
                        f"{error_header}: cannot use {schema_attribute_path.relationship_schema.name} relationship,"
                        f" relationship must be of cardinality many. (`{path}`)"
                    )
                if (
                    not SchemaElementPathType.REL_ONE_OPTIONAL & allowed_path_types
                    and schema_attribute_path.relationship_schema.optional
                    and not (
                        schema_attribute_path.relationship_schema.name == "ip_namespace"
                        and isinstance(node_schema, NodeSchema)
                        and (node_schema.is_ip_address() or node_schema.is_ip_prefix)
                    )
                ):
                    raise ValueError(
                        f"{error_header}: cannot use {schema_attribute_path.relationship_schema.name} relationship,"
                        f" relationship must be mandatory. (`{path}`)"
                    )

            if (
                schema_attribute_path.relationship_schema.cardinality == RelationshipCardinality.MANY
                and not SchemaElementPathType.REL_MANY & allowed_path_types
            ):
                raise ValueError(
                    f"{error_header}: cannot use {schema_attribute_path.relationship_schema.name} relationship,"
                    f" relationship must be of cardinality one (`{path}`)"
                )

            if schema_attribute_path.has_property and not SchemaElementPathType.REL_ATTR & allowed_path_types:
                raise ValueError(
                    f"{error_header}: cannot use attributes of related node, only the relationship. (`{path}`)"
                )
            if not schema_attribute_path.has_property and not SchemaElementPathType.RELS_NO_ATTR & allowed_path_types:
                raise ValueError(f"{error_header}: Must use attributes of related node. (`{path}`)")

        return schema_attribute_path

    def sync_uniqueness_constraints_and_unique_attributes(self) -> None:
        for name in self.generic_names + self.node_names:
            node_schema = self.get(name=name, duplicate=False)

            if not node_schema.unique_attributes and not node_schema.uniqueness_constraints:
                continue

            unique_attrs_in_constraints = set()
            for constraint_paths in node_schema.uniqueness_constraints or []:
                if len(constraint_paths) > 1:
                    continue
                constraint_path = constraint_paths[0]
                try:
                    schema_attribute_path = node_schema.parse_schema_path(path=constraint_path, schema=self)
                except AttributePathParsingError as exc:
                    raise ValueError(
                        f"{node_schema.kind}: Requested unique constraint not found within node. (`{constraint_path}`)"
                    ) from exc

                if (
                    schema_attribute_path.is_type_attribute
                    and schema_attribute_path.attribute_property_name == "value"
                    and schema_attribute_path.attribute_schema
                ):
                    unique_attrs_in_constraints.add(schema_attribute_path.attribute_schema.name)

            unique_attrs_in_attrs = {attr_schema.name for attr_schema in node_schema.unique_attributes}
            if unique_attrs_in_attrs == unique_attrs_in_constraints:
                continue

            attrs_to_make_unique = unique_attrs_in_constraints - unique_attrs_in_attrs
            attrs_to_add_to_constraints = unique_attrs_in_attrs - unique_attrs_in_constraints
            node_schema = self.get(name=name, duplicate=True)

            for attr_name in attrs_to_make_unique:
                attr_schema = node_schema.get_attribute(name=attr_name)
                attr_schema.unique = True
            if attrs_to_add_to_constraints:
                node_schema.uniqueness_constraints = (node_schema.uniqueness_constraints or []) + sorted(
                    [[f"{attr_name}__value"] for attr_name in attrs_to_add_to_constraints]
                )
            self.set(name=name, schema=node_schema)

    def validate_uniqueness_constraints(self) -> None:
        for name in self.all_names:
            node_schema = self.get(name=name, duplicate=False)

            if not node_schema.uniqueness_constraints:
                continue

            for constraint_paths in node_schema.uniqueness_constraints:
                for constraint_path in constraint_paths:
                    element_name = "uniqueness_constraints"
                    self.validate_schema_path(
                        node_schema=node_schema,
                        path=constraint_path,
                        allowed_path_types=SchemaElementPathType.ATTR_WITH_PROP
                        | SchemaElementPathType.REL_ONE_MANDATORY_NO_ATTR,
                        element_name=element_name,
                    )

    def validate_display_labels(self) -> None:
        for name in self.all_names:
            node_schema = self.get(name=name, duplicate=False)

            if node_schema.display_labels:
                for path in node_schema.display_labels:
                    self.validate_schema_path(
                        node_schema=node_schema,
                        path=path,
                        allowed_path_types=SchemaElementPathType.ATTR,
                        element_name="display_labels",
                    )
            elif isinstance(node_schema, NodeSchema):
                generic_display_labels = []
                for generic in node_schema.inherit_from:
                    generic_schema = self.get(name=generic, duplicate=False)
                    if generic_schema.display_labels:
                        generic_display_labels.append(generic_schema.display_labels)

                if len(generic_display_labels) == 1:
                    # Only assign node display labels if a single generic has them defined
                    node_schema.display_labels = generic_display_labels[0]

    def validate_order_by(self) -> None:
        for name in self.all_names:
            node_schema = self.get(name=name, duplicate=False)

            if not node_schema.order_by:
                continue

            allowed_types = SchemaElementPathType.ATTR_WITH_PROP | SchemaElementPathType.REL_ONE_ATTR_WITH_PROP
            for order_by_path in node_schema.order_by:
                element_name = "order_by"
                self.validate_schema_path(
                    node_schema=node_schema,
                    path=order_by_path,
                    allowed_path_types=allowed_types,
                    element_name=element_name,
                )

    def validate_default_filters(self) -> None:
        for name in self.all_names:
            node_schema = self.get(name=name, duplicate=False)

            if not node_schema.default_filter:
                continue

            self.validate_schema_path(
                node_schema=node_schema,
                path=node_schema.default_filter,
                allowed_path_types=SchemaElementPathType.ATTR,
                element_name="default_filter",
            )

    def validate_default_values(self) -> None:
        for name in self.generic_names + self.node_names:
            node_schema = self.get(name=name, duplicate=False)
            for node_attr in node_schema.local_attributes:
                if node_attr.default_value is None:
                    continue

                infrahub_attribute_type = ATTRIBUTE_TYPES[node_attr.kind].get_infrahub_class()
                try:
                    infrahub_attribute_type.validate_format(
                        value=node_attr.default_value, name=node_attr.name, schema=node_attr
                    )
                    infrahub_attribute_type.validate_content(
                        value=node_attr.default_value, name=node_attr.name, schema=node_attr
                    )
                except ValidationError as exc:
                    raise ValidationError(
                        f"{node_schema.namespace}{node_schema.name}: default value {exc.message}"
                    ) from exc

    def validate_human_friendly_id(self) -> None:
        for name in self.generic_names + self.node_names:
            node_schema = self.get(name=name, duplicate=False)
            hf_attr_names = set()

            if not node_schema.human_friendly_id:
                continue

            allowed_types = SchemaElementPathType.ATTR_WITH_PROP | SchemaElementPathType.REL_ONE_MANDATORY_ATTR
            for hfid_path in node_schema.human_friendly_id:
                schema_path = self.validate_schema_path(
                    node_schema=node_schema,
                    path=hfid_path,
                    allowed_path_types=allowed_types,
                    element_name="human_friendly_id",
                )

                if schema_path.is_type_attribute:
                    hf_attr_names.add(schema_path.attribute_schema.name)

    def validate_required_relationships(self) -> None:
        reverse_dependency_map: dict[str, set[str]] = {}
        for name in self.node_names + self.generic_names:
            node_schema = self.get(name=name, duplicate=False)
            for relationship_schema in node_schema.relationships:
                if relationship_schema.optional:
                    continue

                peer_kind = relationship_schema.peer
                if peer_kind in reverse_dependency_map.get(node_schema.kind, set()):
                    raise ValueError(
                        f"'{node_schema.kind}' and '{peer_kind}' cannot both have required relationships to one another."
                    )
                if peer_kind not in reverse_dependency_map:
                    reverse_dependency_map[peer_kind] = set()
                reverse_dependency_map[peer_kind].add(node_schema.kind)

    def validate_parent_component(self) -> None:
        # {parent_kind: {component_kind_1, component_kind_2, ...}}
        dependency_map: dict[str, set[str]] = defaultdict(set)
        for name in self.generic_names + self.node_names:
            node_schema = self.get(name=name, duplicate=False)

            parent_relationships: list[RelationshipSchema] = []
            component_relationships: list[RelationshipSchema] = []
            for rel_schema in node_schema.relationships:
                if rel_schema.kind == RelationshipKind.PARENT and rel_schema.inherited is False:
                    parent_relationships.append(rel_schema)
                    dependency_map[rel_schema.peer].add(node_schema.kind)
                elif rel_schema.kind == RelationshipKind.COMPONENT:
                    component_relationships.append(rel_schema)
                    dependency_map[node_schema.kind].add(rel_schema.peer)

            if isinstance(node_schema, NodeSchema) and node_schema.inherit_from:
                for generic_schema_name in node_schema.inherit_from:
                    generic_schema = self.get_generic(name=generic_schema_name, duplicate=False)
                    generic_parent_relationships = generic_schema.get_relationships_of_kind(
                        relationship_kinds=[RelationshipKind.PARENT]
                    )
                    for gpr in generic_parent_relationships:
                        dependency_map[gpr.peer].add(node_schema.kind)
                    parent_relationships.extend(generic_parent_relationships)
                    generic_component_relationships = generic_schema.get_relationships_of_kind(
                        relationship_kinds=[RelationshipKind.COMPONENT]
                    )
                    for gcr in generic_component_relationships:
                        dependency_map[node_schema.kind].add(gcr.peer)

            if not parent_relationships and not component_relationships:
                continue

            self._validate_parents_one_schema(node_schema=node_schema, parent_relationships=parent_relationships)

        try:
            topological_sort(dependency_map)
        except DependencyCycleExistsError as exc:
            raise ValueError(f"Cycles exist among parents and components in schema: {exc.get_cycle_strings()}") from exc

    def _validate_parents_one_schema(
        self, node_schema: Union[NodeSchema, GenericSchema], parent_relationships: list[RelationshipSchema]
    ) -> None:
        if not parent_relationships:
            return
        if len(parent_relationships) > 1:
            parent_names = [pr.name for pr in parent_relationships]
            raise ValueError(
                f"{node_schema.kind}: Only one relationship of type parent is allowed, but all the following are of type parent: {parent_names}"
            )

        parent_relationship = parent_relationships[0]
        if parent_relationship.cardinality != RelationshipCardinality.ONE:
            raise ValueError(
                f"{node_schema.kind}.{parent_relationship.name}: Relationship of type parent must be cardinality=one"
            )
        if parent_relationship.optional is True:
            raise ValueError(
                f"{node_schema.kind}.{parent_relationship.name}: Relationship of type parent must not be optional"
            )

    def validate_names(self) -> None:
        for name in self.all_names:
            node = self.get(name=name, duplicate=False)

            if names_dup := duplicates(node.attribute_names + node.relationship_names):
                raise ValueError(
                    f"{node.kind}: Names of attributes and relationships must be unique : {names_dup}"
                ) from None

            if node.kind in INTERNAL_SCHEMA_NODE_KINDS:
                continue

            for attr in node.attributes:
                if attr.name in RESERVED_ATTR_REL_NAMES or (
                    isinstance(node, GenericSchema) and attr.name in RESERVED_ATTR_GEN_NAMES
                ):
                    raise ValueError(f"{node.kind}: {attr.name} isn't allowed as an attribute name.")
            for rel in node.relationships:
                if rel.name in RESERVED_ATTR_REL_NAMES or (
                    isinstance(node, GenericSchema) and rel.name in RESERVED_ATTR_GEN_NAMES
                ):
                    raise ValueError(f"{node.kind}: {rel.name} isn't allowed as a relationship name.")

    def validate_kinds(self) -> None:
        for name in list(self.nodes.keys()):
            node = self.get_node(name=name, duplicate=False)

            for generic_kind in node.inherit_from:
                if self.has(name=generic_kind):
                    if not isinstance(self.get(name=generic_kind, duplicate=False), GenericSchema):
                        raise ValueError(
                            f"{node.kind}: Only generic model can be used as part of inherit_from, {generic_kind!r} is not a valid entry."
                        ) from None
                else:
                    raise ValueError(
                        f"{node.kind}: {generic_kind!r} is not a invalid Generic to inherit from"
                    ) from None

            for rel in node.relationships:
                if rel.peer in [InfrahubKind.GENERICGROUP]:
                    continue
                if not self.has(rel.peer):
                    raise ValueError(
                        f"{node.kind}: Relationship {rel.name!r} is referencing an invalid peer {rel.peer!r}"
                    ) from None

    def validate_computed_attributes(self) -> None:
        self.computed_attributes = ComputedAttributes()
        for name in self.nodes.keys():
            node_schema = self.get_node(name=name, duplicate=False)
            for attribute in node_schema.attributes:
                self._validate_computed_attribute(node=node_schema, attribute=attribute)

        for name in self.generics.keys():
            generic_schema = self.get_generic(name=name, duplicate=False)
            for attribute in generic_schema.attributes:
                if attribute.computed_attribute and attribute.computed_attribute.kind != ComputedAttributeKind.USER:
                    raise ValueError(
                        f"{generic_schema.kind}: Attribute {attribute.name!r} computed attributes are only allowed on nodes not generics"
                    )

    def _validate_computed_attribute(self, node: NodeSchema, attribute: AttributeSchema) -> None:
        if not attribute.computed_attribute or attribute.computed_attribute.kind == ComputedAttributeKind.USER:
            return

        if not attribute.read_only:
            raise ValueError(
                f"{node.kind}: Attribute {attribute.name!r} is a computed attribute but not marked as read_only"
            )
        if attribute.kind not in VALID_COMPUTED_ATTRIBUTE_KINDS:
            raise ValueError(
                f"{node.kind}: Attribute {attribute.name!r} is a computed attribute only {VALID_COMPUTED_ATTRIBUTE_KINDS} kinds are supported."
            )

        if (
            attribute.computed_attribute.kind == ComputedAttributeKind.JINJA2
            and not attribute.computed_attribute.jinja2_template
        ):
            raise ValueError(
                f"{node.kind}: Attribute {attribute.name!r} is a computed jinja2 attribute but no logic is defined"
            )
        if (
            attribute.computed_attribute.kind == ComputedAttributeKind.JINJA2
            and attribute.computed_attribute.jinja2_template
        ):
            allowed_path_types = (
                SchemaElementPathType.ATTR_WITH_PROP
                | SchemaElementPathType.REL_ONE_MANDATORY_ATTR_WITH_PROP
                | SchemaElementPathType.REL_ONE_ATTR_WITH_PROP
            )
            try:
                macro = MacroDefinition(macro=attribute.computed_attribute.jinja2_template)
            except ValueError as exc:
                raise ValueError(
                    f"{node.kind}: Attribute {attribute.name!r} is assigned by a jinja2 template, but has an invalid template"
                ) from exc

            for variable in macro.variables:
                try:
                    schema_path = self.validate_schema_path(
                        node_schema=node, path=variable, allowed_path_types=allowed_path_types
                    )
                except ValueError as exc:
                    raise ValueError(
                        f"{node.kind}: Attribute {attribute.name!r} the '{variable}' variable is not found within the schema path"
                    ) from exc

                if schema_path.is_type_attribute and schema_path.active_attribute_schema.name == attribute.name:
                    raise ValueError(
                        f"{node.kind}: Attribute {attribute.name!r} the '{variable}' variable is a reference to itself"
                    )

                self.computed_attributes.register_computed_jinja2(
                    node=node, attribute=attribute, schema_path=schema_path
                )

        elif attribute.computed_attribute.kind == ComputedAttributeKind.TRANSFORM_PYTHON and not attribute.optional:
            raise ValueError(
                f"{node.kind}: Attribute {attribute.name!r} is a computed transform, it can't be mandatory"
            )

        elif attribute.computed_attribute.kind == ComputedAttributeKind.TRANSFORM_PYTHON:
            self.computed_attributes.add_python_attribute(node=node, attribute=attribute)

    def validate_count_against_cardinality(self) -> None:
        """Validate every RelationshipSchema cardinality against the min_count and max_count."""
        for name in self.all_names:
            node = self.get(name=name, duplicate=False)

            for rel in node.relationships:
                if rel.cardinality == RelationshipCardinality.ONE:
                    if not rel.optional and (rel.min_count != 1 or rel.max_count != 1):
                        raise ValueError(
                            f"{node.kind}: Relationship {rel.name!r} is defined as cardinality.ONE but min_count or max_count are not 1"
                        ) from None
                elif rel.cardinality == RelationshipCardinality.MANY:
                    if rel.max_count and rel.min_count > rel.max_count:
                        raise ValueError(
                            f"{node.kind}: Relationship {rel.name!r} min_count must be lower than max_count"
                        )
                    if rel.max_count == 1:
                        raise ValueError(
                            f"{node.kind}: Relationship {rel.name!r} max_count must be 0 or greater than 1 when cardinality is MANY"
                        )

    def process_dropdowns(self) -> None:
        for name in self.all_names:
            node = self.get(name=name, duplicate=False)

            attributes = [attr for attr in node.attributes if attr.kind == "Dropdown"]
            if not attributes:
                continue

            node = node.duplicate()
            changed = False

            for attr in attributes:
                if not attr.choices:
                    continue

                sorted_choices = sorted(attr.choices or [], key=lambda x: x.name, reverse=True)
                defined_colors = [choice.color for choice in sorted_choices if choice.color]
                for choice in sorted_choices:
                    if not choice.color:
                        choice.color = select_color(defined_colors)
                    if not choice.label:
                        choice.label = format_label(choice.name)
                    if not choice.description:
                        choice.description = ""

                if attr.choices != sorted_choices:
                    attr.choices = sorted_choices
                    changed = True

            if changed:
                self.set(name=name, schema=node)

    def process_labels(self) -> None:
        def check_if_need_to_update_label(node) -> bool:
            if not node.label:
                return True
            for item in node.relationships + node.attributes:
                if not item.label:
                    return True
            return False

        for name in self.all_names:
            node = self.get(name=name, duplicate=False)

            if not check_if_need_to_update_label(node):
                continue

            node = node.duplicate()
            if not node.label:
                node.label = format_label(node.name)

            for attr in node.attributes:
                if not attr.label:
                    attr.label = format_label(attr.name)

            for rel in node.relationships:
                if not rel.label:
                    rel.label = format_label(rel.name)

            self.set(name=name, schema=node)

    def process_relationships(self) -> None:
        for name in self.all_names:
            node = self.get(name=name, duplicate=False)

            schema_to_update: Optional[Union[NodeSchema, GenericSchema]] = None
            for relationship in node.relationships:
                if relationship.on_delete is not None:
                    continue
                if not schema_to_update:
                    schema_to_update = node.duplicate()

                relationship_to_update = schema_to_update.get_relationship(name=relationship.name)
                if relationship.kind == RelationshipKind.COMPONENT:
                    relationship_to_update.on_delete = RelationshipDeleteBehavior.CASCADE
                else:
                    relationship_to_update.on_delete = RelationshipDeleteBehavior.NO_ACTION

            if schema_to_update:
                self.set(name=schema_to_update.kind, schema=schema_to_update)

    def process_human_friendly_id(self) -> None:
        for name in self.generic_names + self.node_names:
            node = self.get(name=name, duplicate=False)

            # If human_friendly_id IS NOT defined
            #   but some the model has some unique attribute, we generate a human_friendly_id
            # If human_friendly_id IS defined
            #   but no unique attributes and no uniquess constraints, we add a uniqueness_constraint
            if not node.human_friendly_id and node.unique_attributes:
                for attr in node.unique_attributes:
                    node = self.get(name=name, duplicate=True)
                    node.human_friendly_id = [f"{attr.name}__value"]
                    self.set(name=node.kind, schema=node)
                    break
                continue

            # if no human_friendly_id and a uniqueness_constraint with a single attribute exists
            # then use that attribute as the human_friendly_id
            if not node.human_friendly_id and node.uniqueness_constraints:
                for constraint_paths in node.uniqueness_constraints:
                    if len(constraint_paths) > 1:
                        continue
                    constraint_path = constraint_paths[0]
                    schema_path = node.parse_schema_path(path=constraint_path, schema=node)
                    if (
                        schema_path.is_type_attribute
                        and schema_path.attribute_property_name == "value"
                        and schema_path.attribute_schema
                    ):
                        node = self.get(name=name, duplicate=True)
                        node.human_friendly_id = [f"{schema_path.attribute_schema.name}__value"]
                        self.set(name=node.kind, schema=node)
                        break

            if node.human_friendly_id and not node.uniqueness_constraints:
                uniqueness_constraints: list[str] = []
                for item in node.human_friendly_id:
                    schema_attribute_path = node.parse_schema_path(path=item, schema=self)
                    if schema_attribute_path.is_type_attribute:
                        uniqueness_constraints.append(item)
                    elif schema_attribute_path.is_type_relationship:
                        uniqueness_constraints.append(schema_attribute_path.relationship_schema.name)

                node.uniqueness_constraints = [uniqueness_constraints]
                self.set(name=node.kind, schema=node)

    def process_hierarchy(self) -> None:
        for name in self.nodes.keys():
            node = self.get_node(name=name, duplicate=False)

            if not node.hierarchy and not node.parent and not node.children:
                continue

            if not node.hierarchy and (node.parent is not None or node.children is not None):
                raise ValueError(f"{node.kind} Hierarchy must be provided if either parent or children is defined.")

            if node.hierarchy not in self.generics.keys():
                # TODO add a proper exception for all schema related issue
                raise ValueError(f"{node.kind} Unable to find the generic {node.hierarchy!r} provided in 'hierarchy'.")

            node = node.duplicate()
            changed = False

            if node.hierarchy not in node.inherit_from:
                node.inherit_from.append(node.hierarchy)
                changed = True

            if node.parent is None:
                node.parent = node.hierarchy
                changed = True
            elif node.parent and node.parent not in list(self.nodes.keys()) + list(self.generics.keys()):
                raise ValueError(f"{node.kind} Unable to find the node {node.parent!r} provided in 'parent'.")

            if node.children is None:
                node.children = node.hierarchy
                changed = True
            elif node.children and node.children not in list(self.nodes.keys()) + list(self.generics.keys()):
                raise ValueError(f"{node.kind} Unable to find the node {node.children!r} provided in 'children'.")

            if changed:
                self.set(name=name, schema=node)

    def process_inheritance(self) -> None:
        """Extend all the nodes with the attributes and relationships
        from the Interface objects defined in inherited_from.
        """

        generics_used_by = defaultdict(list)

        # For all node_schema, add the attributes & relationships from the generic / interface
        for name in self.nodes.keys():
            node = self.get_node(name=name, duplicate=False)

            if node.inherit_from or node.namespace not in RESTRICTED_NAMESPACES:
                generics_used_by[InfrahubKind.NODE].append(node.kind)

            if not node.inherit_from:
                continue

            node = node.duplicate()

            if InfrahubKind.IPPREFIX in node.inherit_from and InfrahubKind.IPADDRESS in node.inherit_from:
                raise ValueError(
                    f"{node.kind} cannot inherit from both {InfrahubKind.IPPREFIX} and {InfrahubKind.IPADDRESS}"
                )

            generic_with_hierarchical_support = []
            for generic_kind in node.inherit_from:
                if generic_kind not in self.generics.keys():
                    # TODO add a proper exception for all schema related issue
                    raise ValueError(f"{node.kind} Unable to find the generic {generic_kind}")

                generic_kind_schema = self.get_generic(generic_kind, duplicate=False)
                if generic_kind_schema.hierarchical:
                    generic_with_hierarchical_support.append(generic_kind)

                # Perform checks to validate that the node is not breaking inheritance rules
                node.validate_inheritance(interface=generic_kind_schema)

                # Store the list of node referencing a specific generics
                generics_used_by[generic_kind].append(node.kind)
                node.inherit_from_interface(interface=generic_kind_schema)

            if len(generic_with_hierarchical_support) > 1:
                raise ValueError(
                    f"{node.kind} Only one generic with hierarchical support is allowed per node {generic_with_hierarchical_support}"
                )
            if len(generic_with_hierarchical_support) == 1 and node.hierarchy is None:
                node.hierarchy = generic_with_hierarchical_support[0]

            self.set(name=name, schema=node)

        # Update all generics with the list of nodes referrencing them.
        for generic_name in self.generics.keys():
            generic = self.get(name=generic_name)

            if generic.kind in generics_used_by:
                generic.used_by = sorted(generics_used_by[generic.kind])
            else:
                generic.used_by = []

            self.set(name=generic_name, schema=generic)

    def process_branch_support(self) -> None:
        """Set branch support flag on all attributes and relationships if not already defined.

        if either node on a relationship support branch, the relationship must be branch aware.
        """
        # pylint: disable=too-many-branches

        for name in self.all_names:
            node = self.get(name=name, duplicate=False)

            # Check if this node requires a change before duplicating
            change_required = False
            for attr in node.attributes:
                if attr.branch is None:
                    change_required = True
                    break
            if not change_required:
                for rel in node.relationships:
                    if rel.branch is None:
                        change_required = True
                        break

            if not change_required:
                continue

            node = node.duplicate()

            for attr in node.attributes:
                if attr.branch is not None:
                    continue

                attr.branch = node.branch

            for rel in node.relationships:
                if rel.branch is not None:
                    continue

                peer_node = self.get(name=rel.peer, duplicate=False)
                if node.branch == peer_node.branch:
                    rel.branch = node.branch
                elif BranchSupportType.LOCAL in (node.branch, peer_node.branch):
                    rel.branch = BranchSupportType.LOCAL
                else:
                    rel.branch = BranchSupportType.AWARE

            self.set(name=name, schema=node)

    def process_default_values(self) -> None:
        """Ensure that all attributes with a default value are flagged as optional: True."""
        for name in self.all_names:
            node = self.get(name=name, duplicate=False)

            attr_names_to_update = [
                attr.name for attr in node.attributes if attr.default_value is not None and not attr.optional
            ]
            if not attr_names_to_update:
                continue

            node = node.duplicate()
            for attr_name in attr_names_to_update:
                attr = node.get_attribute(name=attr_name)
                attr.optional = True

            self.set(name=name, schema=node)

    def process_cardinality_counts(self) -> None:
        """Ensure that all relationships with a cardinality of ONE have a min_count and max_count of 1."""
        # pylint: disable=too-many-branches

        for name in self.all_names:
            node = self.get(name=name, duplicate=False)

            # Check if this node requires a change before duplicating
            change_required = False
            for rel in node.relationships:
                if rel.cardinality != RelationshipCardinality.ONE:
                    continue
                # Handle default values of RelationshipSchema when cardinality is ONE and set to valid values (1)
                # RelationshipSchema default values 0 for min_count and max_count
                if rel.optional and rel.min_count != 0:
                    change_required = True
                    break
                if rel.optional and rel.max_count != 1:
                    change_required = True
                    break
                if not rel.optional and rel.min_count == 0:
                    change_required = True
                    break
                if not rel.optional and rel.max_count == 0:
                    change_required = True
                    break

            if not change_required:
                continue

            node = node.duplicate()

            for rel in node.relationships:
                if rel.cardinality != RelationshipCardinality.ONE:
                    continue
                # Handle default values of RelationshipSchema when cardinality is ONE and set to valid values (1)
                # RelationshipSchema default values 0 for min_count and max_count
                if rel.optional and rel.min_count != 0:
                    rel.min_count = 0
                if rel.optional and rel.max_count != 1:
                    rel.max_count = 1
                if not rel.optional and rel.min_count == 0:
                    rel.min_count = 1
                if not rel.optional and rel.max_count == 0:
                    rel.max_count = 1

            self.set(name=name, schema=node)

    def process_deprecations(self) -> None:
        """Mark deprecated attributes and relationships as optional."""
        for name in self.all_names:
            node = self.get(name=name, duplicate=False)

            change_required = False
            for item in node.attributes + node.relationships:
                if item.is_deprecated:
                    log.warn(f"'{item.name}' for '{node.kind}' has been marked as deprecated, remember to clean it up")
                    if not item.optional:
                        change_required = True

            if not change_required:
                continue

            node = node.duplicate()

            for item in node.attributes + node.relationships:
                if item.is_deprecated and not item.optional:
                    item.optional = True

            self.set(name=name, schema=node)

    def generate_weight(self) -> None:
        for name in self.all_names:
            node = self.get(name=name, duplicate=False)
            items_to_update = [item for item in node.attributes + node.relationships if not item.order_weight]
            if not items_to_update:
                continue

            node = node.duplicate()

            current_weight = 0
            for item in node.attributes + node.relationships:
                current_weight += 1000
                if not item.order_weight:
                    item.order_weight = current_weight

            self.set(name=name, schema=node)

    def cleanup_inherited_elements(self) -> None:
        # pylint: disable=too-many-branches
        for name in self.node_names:
            node = self.get_node(name=name, duplicate=False)

            attributes_to_delete = []
            relationships_to_delete = []

            inherited_attribute_names = set(node.attribute_names) - set(node.local_attribute_names)
            inherited_relationship_names = set(node.relationship_names) - set(node.local_relationship_names)
            for item_name in inherited_attribute_names:
                found = False
                for generic_name in node.inherit_from:
                    generic = self.get_generic(name=generic_name, duplicate=False)
                    if item_name in generic.attribute_names:
                        attr = generic.get_attribute(name=item_name)
                        if attr.state != HashableModelState.ABSENT:
                            found = True
                if not found:
                    attributes_to_delete.append(item_name)

            for item_name in inherited_relationship_names:
                found = False
                for generic_name in node.inherit_from:
                    generic = self.get_generic(name=generic_name, duplicate=False)
                    if item_name in generic.relationship_names:
                        rel = generic.get_relationship(name=item_name)
                        if rel.state != HashableModelState.ABSENT:
                            found = True
                if not found:
                    relationships_to_delete.append(item_name)

            # If there is either an attribute or a relationship to delete
            # We clone the node and we set the attribute / relationship as ABSENT
            if attributes_to_delete or relationships_to_delete:
                node_copy = self.get_node(name=name, duplicate=True)
                for item_name in attributes_to_delete:
                    attr = node_copy.get_attribute(name=item_name)
                    attr.state = HashableModelState.ABSENT
                for item_name in relationships_to_delete:
                    rel = node_copy.get_relationship(name=item_name)
                    rel.state = HashableModelState.ABSENT
                self.set(name=name, schema=node_copy)

    def add_groups(self) -> None:
        if not self.has(name=InfrahubKind.GENERICGROUP):
            return

        for node_name in self.all_names:
            schema: MainSchemaTypes = self.get(name=node_name, duplicate=False)
            changed = False

            if isinstance(schema, NodeSchema) and InfrahubKind.GENERICGROUP in schema.inherit_from:
                continue

            if schema.kind in INTERNAL_SCHEMA_NODE_KINDS or schema.kind == InfrahubKind.GENERICGROUP:
                continue

            if schema.kind in (InfrahubKind.LINEAGEOWNER, InfrahubKind.LINEAGESOURCE):
                continue

            if "member_of_groups" not in schema.relationship_names:
                if not changed:
                    schema = schema.duplicate()
                    changed = True
                schema.relationships.append(
                    RelationshipSchema(
                        name="member_of_groups",
                        identifier="group_member",
                        peer=InfrahubKind.GENERICGROUP,
                        kind=RelationshipKind.GROUP,
                        cardinality=RelationshipCardinality.MANY,
                        branch=BranchSupportType.AWARE,
                    )
                )

            if "subscriber_of_groups" not in schema.relationship_names:
                if not changed:
                    schema = schema.duplicate()
                    changed = True
                schema.relationships.append(
                    RelationshipSchema(
                        name="subscriber_of_groups",
                        identifier="group_subscriber",
                        peer=InfrahubKind.GENERICGROUP,
                        kind=RelationshipKind.GROUP,
                        cardinality=RelationshipCardinality.MANY,
                        branch=BranchSupportType.AWARE,
                    )
                )

            if changed:
                self.set(name=node_name, schema=schema)

    def _get_hierarchy_child_rel(self, peer: str, hierarchical: str, read_only: bool) -> RelationshipSchema:
        return RelationshipSchema(
            name="children",
            identifier="parent__child",
            peer=peer,
            kind=RelationshipKind.HIERARCHY,
            cardinality=RelationshipCardinality.MANY,
            branch=BranchSupportType.AWARE,
            direction=RelationshipDirection.INBOUND,
            hierarchical=hierarchical,
            read_only=read_only,
        )

    def _get_hierarchy_parent_rel(self, peer: str, hierarchical: str, read_only: bool) -> RelationshipSchema:
        return RelationshipSchema(
            name="parent",
            identifier="parent__child",
            peer=peer,
            kind=RelationshipKind.HIERARCHY,
            cardinality=RelationshipCardinality.ONE,
            max_count=1,
            branch=BranchSupportType.AWARE,
            direction=RelationshipDirection.OUTBOUND,
            hierarchical=hierarchical,
            read_only=read_only,
        )

    def add_hierarchy_generic(self) -> None:
        for generic_name in self.generics.keys():
            generic = self.get_generic(name=generic_name, duplicate=False)

            if not generic.hierarchical:
                continue

            generic = generic.duplicate()
            read_only = generic.kind == InfrahubKind.IPPREFIX

            if "parent" not in generic.relationship_names:
                generic.relationships.append(
                    self._get_hierarchy_parent_rel(peer=generic_name, hierarchical=generic_name, read_only=read_only)
                )
            if "children" not in generic.relationship_names:
                generic.relationships.append(
                    self._get_hierarchy_child_rel(peer=generic_name, hierarchical=generic_name, read_only=read_only)
                )

            self.set(name=generic_name, schema=generic)

    def add_hierarchy_node(self) -> None:
        for node_name in self.nodes.keys():
            node = self.get_node(name=node_name, duplicate=False)

            if node.parent is None and node.children is None:
                continue

            node = node.duplicate()
            read_only = InfrahubKind.IPPREFIX in node.inherit_from

            if node.parent:
                if "parent" not in node.relationship_names:
                    node.relationships.append(
                        self._get_hierarchy_parent_rel(
                            peer=node.parent, hierarchical=node.hierarchy, read_only=read_only
                        )
                    )
                else:
                    parent_rel = node.get_relationship(name="parent")
                    if parent_rel.peer != node.parent:
                        parent_rel.peer = node.parent

            if node.children:
                if "children" not in node.relationship_names:
                    node.relationships.append(
                        self._get_hierarchy_child_rel(
                            peer=node.children, hierarchical=node.hierarchy, read_only=read_only
                        )
                    )
                else:
                    children_rel = node.get_relationship(name="children")
                    if children_rel.peer != node.children:
                        children_rel.peer = node.children

            self.set(name=node_name, schema=node)

    def manage_profile_schemas(self) -> None:
        if not self.has(name=InfrahubKind.PROFILE):
            # TODO: This logic is actually only for testing purposes as since 1.0.9 CoreProfile is loaded in db.
            #  Ideally, we would remove this and instead load CoreProfile properly within tests.
            core_profile_schema = GenericSchema(**core_profile_schema_definition)
            self.set(name=core_profile_schema.kind, schema=core_profile_schema)

        profile_schema_kinds = set()
        for node_name in self.node_names + self.generic_names:
            node = self.get(name=node_name, duplicate=False)
            if (
                node.namespace in RESTRICTED_NAMESPACES
                or not node.generate_profile
                or node.state == HashableModelState.ABSENT
            ):
                try:
                    self.delete(name=self._get_profile_kind(node_kind=node.kind))
                except SchemaNotFoundError:
                    ...
                continue

            profile = self.generate_profile_from_node(node=node)
            self.set(name=profile.kind, schema=profile)
            profile_schema_kinds.add(profile.kind)

        for previous_profile in list(self.profiles.keys()):
            # Ensure that we remove previous profile schemas if a node has been renamed
            if previous_profile not in profile_schema_kinds:
                self.delete(name=previous_profile)

        if not profile_schema_kinds:
            return

        # Update used_by list for CoreProfile and CoreNode
        core_profile_schema = self.get(name=InfrahubKind.PROFILE, duplicate=False)
        current_used_by_profile = set(core_profile_schema.used_by)
        new_used_by_profile = profile_schema_kinds - current_used_by_profile

        if new_used_by_profile:
            core_profile_schema = self.get(name=InfrahubKind.PROFILE, duplicate=True)
            core_profile_schema.used_by = sorted(list(profile_schema_kinds))
            self.set(name=InfrahubKind.PROFILE, schema=core_profile_schema)

        if self.has(name=InfrahubKind.NODE):
            core_node_schema = self.get(name=InfrahubKind.NODE, duplicate=False)
            current_used_by_node = set(core_node_schema.used_by)
            new_used_by_node = profile_schema_kinds - current_used_by_node

            if new_used_by_node:
                core_node_schema = self.get(name=InfrahubKind.NODE, duplicate=True)
                updated_used_by_node = set(chain(profile_schema_kinds, set(core_node_schema.used_by)))
                core_node_schema.used_by = sorted(list(updated_used_by_node))
                self.set(name=InfrahubKind.NODE, schema=core_node_schema)

    def manage_profile_relationships(self) -> None:
        for node_name in self.node_names + self.generic_names:
            node = self.get(name=node_name, duplicate=False)

            if node.namespace in RESTRICTED_NAMESPACES:
                continue

            profiles_rel_settings: dict[str, Any] = {
                "name": "profiles",
                "identifier": "node__profile",
                "peer": InfrahubKind.PROFILE,
                "kind": RelationshipKind.PROFILE,
                "cardinality": RelationshipCardinality.MANY,
                "branch": BranchSupportType.AWARE,
            }

            # Add relationship between node and profile
            if "profiles" not in node.relationship_names:
                node_schema = self.get(name=node_name, duplicate=True)

                node_schema.relationships.append(RelationshipSchema(**profiles_rel_settings))
                self.set(name=node_name, schema=node_schema)
            else:
                has_changes: bool = False
                rel_profiles = node.get_relationship(name="profiles")
                for name, value in profiles_rel_settings.items():
                    if getattr(rel_profiles, name) != value:
                        has_changes = True

                if not has_changes:
                    continue

                node_schema = self.get(name=node_name, duplicate=True)
                rel_profiles = node_schema.get_relationship(name="profiles")
                for name, value in profiles_rel_settings.items():
                    if getattr(rel_profiles, name) != value:
                        setattr(rel_profiles, name, value)

                self.set(name=node_name, schema=node_schema)

    def _get_profile_kind(self, node_kind: str) -> str:
        return f"Profile{node_kind}"

    def generate_profile_from_node(self, node: NodeSchema) -> ProfileSchema:
        core_profile_schema = self.get(name=InfrahubKind.PROFILE, duplicate=False)
        core_name_attr = core_profile_schema.get_attribute(name="profile_name")
        profile_name_attr = AttributeSchema(
            **core_name_attr.model_dump(exclude=["id", "inherited"]),
        )
        profile_name_attr.branch = node.branch
        core_priority_attr = core_profile_schema.get_attribute(name="profile_priority")
        profile_priority_attr = AttributeSchema(
            **core_priority_attr.model_dump(exclude=["id", "inherited"]),
        )
        profile_priority_attr.branch = node.branch
        profile = ProfileSchema(
            name=node.kind,
            namespace="Profile",
            label=f"Profile {node.label}",
            description=f"Profile for {node.kind}",
            branch=node.branch,
            include_in_menu=False,
            display_labels=["profile_name__value"],
            inherit_from=[InfrahubKind.LINEAGESOURCE, InfrahubKind.PROFILE, InfrahubKind.NODE],
            human_friendly_id=["profile_name__value"],
            default_filter="profile_name__value",
            attributes=[profile_name_attr, profile_priority_attr],
            relationships=[
                RelationshipSchema(
                    name="related_nodes",
                    identifier="node__profile",
                    peer=node.kind,
                    kind=RelationshipKind.PROFILE,
                    cardinality=RelationshipCardinality.MANY,
                    branch=BranchSupportType.AWARE,
                )
            ],
        )

        for node_attr in node.attributes:
            if node_attr.read_only or node_attr.optional is False:
                continue

            attr = AttributeSchema(
                optional=True,
                **node_attr.model_dump(exclude=["id", "unique", "optional", "read_only", "default_value", "inherited"]),
            )
            profile.attributes.append(attr)

        return profile



from __future__ import annotations

from typing import TYPE_CHECKING, Optional, Union

from infrahub.core.constants import RepositoryInternalStatus
from infrahub.core.diff.model.path import BranchTrackingId
from infrahub.core.manager import NodeManager
from infrahub.core.models import SchemaUpdateValidationResult
from infrahub.core.protocols import CoreRepository
from infrahub.core.registry import registry
from infrahub.core.timestamp import Timestamp
from infrahub.exceptions import ValidationError
from infrahub.log import get_logger

from ..git.models import GitRepositoryMerge
from ..workflows.catalogue import GIT_REPOSITORIES_MERGE

if TYPE_CHECKING:
    from infrahub.core.branch import Branch
    from infrahub.core.diff.coordinator import DiffCoordinator
    from infrahub.core.diff.merger.merger import DiffMerger
    from infrahub.core.diff.repository.repository import DiffRepository
    from infrahub.core.models import SchemaUpdateConstraintInfo, SchemaUpdateMigrationInfo
    from infrahub.core.schema.manager import SchemaDiff
    from infrahub.core.schema.schema_branch import SchemaBranch
    from infrahub.database import InfrahubDatabase
    from infrahub.services import InfrahubServices

log = get_logger()


class BranchMerger:
    def __init__(
        self,
        db: InfrahubDatabase,
        source_branch: Branch,
        diff_coordinator: DiffCoordinator,
        diff_merger: DiffMerger,
        diff_repository: DiffRepository,
        destination_branch: Optional[Branch] = None,
        service: Optional[InfrahubServices] = None,
    ):
        self.source_branch = source_branch
        self.destination_branch: Branch = destination_branch or registry.get_branch_from_registry()
        self.db = db
        self.diff_coordinator = diff_coordinator
        self.diff_merger = diff_merger
        self.diff_repository = diff_repository
        self.migrations: list[SchemaUpdateMigrationInfo] = []
        self._merge_at = Timestamp()

        self._source_schema: Optional[SchemaBranch] = None
        self._destination_schema: Optional[SchemaBranch] = None
        self._initial_source_schema: Optional[SchemaBranch] = None

        self._service = service

    @property
    def source_schema(self) -> SchemaBranch:
        if not self._source_schema:
            self._source_schema = registry.schema.get_schema_branch(name=self.source_branch.name).duplicate()

        return self._source_schema

    @property
    def destination_schema(self) -> SchemaBranch:
        if not self._destination_schema:
            self._destination_schema = registry.schema.get_schema_branch(name=self.destination_branch.name).duplicate()

        return self._destination_schema

    @property
    def initial_source_schema(self) -> SchemaBranch:
        if self._initial_source_schema:
            return self._initial_source_schema
        raise ValueError("_initial_source_schema hasn't been initialized")

    @property
    def service(self) -> InfrahubServices:
        if not self._service:
            raise ValueError("BranchMerger hasn't been initialized with a service object")
        return self._service

    async def get_initial_source_branch(self) -> SchemaBranch:
        """Retrieve the schema of the source branch when the branch was created.
        For now we are querying the full schema, but this is something we'll need to revisit in the future by either:
         - having a faster way to pull a previous version of the schema
         - using the diff generated from the data
        """
        if self._initial_source_schema:
            return self._initial_source_schema

        self._initial_source_schema = await registry.schema.load_schema_from_db(
            db=self.db,
            branch=self.source_branch,
            at=Timestamp(self.source_branch.created_at),
        )

        return self._initial_source_schema

    async def has_schema_changes(self) -> bool:
        diff_summary = await self.diff_repository.summary(
            base_branch_name=self.destination_branch.name,
            diff_branch_names=[self.source_branch.name],
            tracking_id=BranchTrackingId(name=self.source_branch.name),
            filters={"kind": {"includes": ["SchemaNode", "SchemaAttribute", "SchemaRelationship"]}},
        )
        if not diff_summary:
            return False
        return bool(diff_summary.num_added or diff_summary.num_removed or diff_summary.num_updated)

    async def update_schema(self) -> bool:
        """After the merge, if there was some changes, we need to:
        - update the schema in the registry
        - Identify if we need to execute some migrations
        """

        # NOTE we need to revisit how to calculate an accurate diff to pull only what needs to be updated from the schema
        # for now the best solution is to pull everything to ensure the integrity of the schema

        if not await self.has_schema_changes():
            return False

        updated_schema = await registry.schema.load_schema_from_db(
            db=self.db,
            branch=self.destination_branch,
            # schema=self.destination_schema.duplicate(),
            # schema_diff=schema_diff,
        )
        registry.schema.set_schema_branch(name=self.destination_branch.name, schema=updated_schema)
        self.destination_branch.update_schema_hash()
        await self.destination_branch.save(db=self.db)

        await self.calculate_migrations(target_schema=updated_schema)

        return True

    def get_candidate_schema(self) -> SchemaBranch:
        # For now, we retrieve the latest schema for each branch from the registry
        # In the future it would be good to generate the object SchemaUpdateValidationResult from message.branch_diff
        current_schema = self.source_schema.duplicate()
        candidate_schema = self.destination_schema.duplicate()
        candidate_schema.update(schema=current_schema)

        return candidate_schema

    async def get_3ways_diff_schema(self) -> SchemaDiff:
        # To calculate the migrations that we need to execute we need
        # the initial version of the schema when the branch was created
        # and we need to calculate a 3 ways comparison between
        # - The initial schema and the current schema in the source branch
        # - The initial schema and the current schema in the destination branch
        initial_source_schema = await self.get_initial_source_branch()

        diff_source = initial_source_schema.diff(other=self.source_schema)
        diff_destination = initial_source_schema.diff(other=self.destination_schema)
        diff_both = diff_source + diff_destination

        return diff_both

    async def calculate_migrations(self, target_schema: SchemaBranch) -> list[SchemaUpdateMigrationInfo]:
        diff_3way = await self.get_3ways_diff_schema()
        validation = SchemaUpdateValidationResult.init(diff=diff_3way, schema=target_schema)
        self.migrations = validation.migrations
        return self.migrations

    async def calculate_validations(self, target_schema: SchemaBranch) -> list[SchemaUpdateConstraintInfo]:
        diff_3way = await self.get_3ways_diff_schema()
        validation = SchemaUpdateValidationResult.init(diff=diff_3way, schema=target_schema)
        return validation.constraints

    async def merge(
        self,
        at: Optional[Union[str, Timestamp]] = None,
    ) -> None:
        """Merge the current branch into main."""
        if self.source_branch.name == registry.default_branch:
            raise ValidationError(f"Unable to merge the branch '{self.source_branch.name}' into itself")

        log.debug("Updating diff for merge")
        enriched_diff = await self.diff_coordinator.update_branch_diff_and_return(
            base_branch=self.destination_branch, diff_branch=self.source_branch
        )
        log.debug("Diff updated for merge")
        conflict_map = enriched_diff.get_all_conflicts()
        errors: list[str] = []
        for conflict_path, conflict in conflict_map.items():
            if conflict.selected_branch is None or conflict.resolvable is False:
                errors.append(conflict_path)

        if errors:
            raise ValidationError(
                f"Unable to merge the branch '{self.source_branch.name}', conflict resolution missing: {', '.join(errors)}"
            )

        # TODO need to find a way to properly communicate back to the user any issue that could come up during the merge
        # From the Graph or From the repositories
        self._merge_at = Timestamp(at)
        await self.diff_merger.merge_graph(at=self._merge_at)
        await self.merge_repositories()

    async def rollback(self) -> None:
        await self.diff_merger.rollback(at=self._merge_at)

    async def merge_repositories(self) -> None:
        # Collect all Repositories in Main because we'll need the commit in Main for each one.
        repos_in_main_list = await NodeManager.query(schema=CoreRepository, db=self.db)
        repos_in_main = {repo.id: repo for repo in repos_in_main_list}

        repos_in_branch_list = await NodeManager.query(schema=CoreRepository, db=self.db, branch=self.source_branch)
        for repo in repos_in_branch_list:
            # Check if the repo, exist in main, if not ignore this repo
            if repo.id not in repos_in_main:
                continue

            if repo.internal_status.value == RepositoryInternalStatus.INACTIVE.value:
                continue

            if self.source_branch.sync_with_git or repo.internal_status.value == RepositoryInternalStatus.STAGING.value:
                model = GitRepositoryMerge(
                    repository_id=repo.id,
                    repository_name=repo.name.value,
                    internal_status=repo.internal_status.value,
                    source_branch=self.source_branch.name,
                    destination_branch=self.destination_branch.name,
                    destination_branch_id=str(self.destination_branch.get_uuid()),
                    default_branch=repo.default_branch.value,
                )
                await self.service.workflow.submit_workflow(
                    workflow=GIT_REPOSITORIES_MERGE, parameters={"model": model}
                )



from infrahub.core import registry
from infrahub.core.diff.merger.merger import DiffMerger
from infrahub.core.diff.merger.serializer import DiffMergeSerializer
from infrahub.dependencies.interface import DependencyBuilder, DependencyBuilderContext

from .repository import DiffRepositoryDependency


class DiffMergerDependency(DependencyBuilder[DiffMerger]):
    @classmethod
    def build(cls, context: DependencyBuilderContext) -> DiffMerger:
        return DiffMerger(
            db=context.db,
            source_branch=context.branch,
            destination_branch=registry.get_branch_from_registry(),
            diff_repository=DiffRepositoryDependency.build(context=context),
            serializer=DiffMergeSerializer(db=context.db, max_batch_size=100),
        )



from pydantic import Field

from infrahub.message_bus import InfrahubMessage


class EventBranchRebased(InfrahubMessage):
    """Sent when a branch has been rebased."""

    branch: str = Field(..., description="The branch that was rebased")



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional

from infrahub.core.constants import PathType
from infrahub.core.path import DataPath, GroupedDataPaths

from ..interface import ConstraintCheckerInterface
from ..shared import (
    SchemaValidatorQuery,
)

if TYPE_CHECKING:
    from infrahub.core.branch import Branch
    from infrahub.database import InfrahubDatabase

    from ..model import SchemaConstraintValidatorRequest


class NodeGenerateProfileValidatorQuery(SchemaValidatorQuery):
    name = "node_constraints_generate_profile_validator"

    def __init__(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.profile_kind = f"Profile{self.node_schema.kind}"

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:
        branch_filter, branch_params = self.branch.get_query_filter_path(at=self.at)
        self.params.update(branch_params)

        query = """
            MATCH (n:%(profile_kind)s)-[r:IS_PART_OF]->(:Root)
            WHERE %(branch_filter)s
            WITH n, r
            ORDER BY n.uuid, r.branch_level DESC, r.from DESC
            WITH n, head(collect(r)) AS latest_r
            WHERE latest_r.status = "active"
        """ % {"profile_kind": self.profile_kind, "branch_filter": branch_filter}
        self.add_to_query(query)
        self.return_labels = ["latest_r.branch AS branch_name", "n.uuid AS node_uuid"]

    async def get_paths(self) -> GroupedDataPaths:
        grouped_data_paths = GroupedDataPaths()
        for result in self.results:
            grouped_data_paths.add_data_path(
                DataPath(
                    branch=str(result.get("branch_name")),
                    path_type=PathType.NODE,
                    node_id=str(result.get("node_uuid")),
                    kind=self.profile_kind,
                )
            )

        return grouped_data_paths


class NodeGenerateProfileChecker(ConstraintCheckerInterface):
    query_classes = [NodeGenerateProfileValidatorQuery]

    def __init__(self, db: InfrahubDatabase, branch: Optional[Branch] = None) -> None:
        self.db = db
        self.branch = branch

    @property
    def name(self) -> str:
        return "node.generate_profile.update"

    def supports(self, request: SchemaConstraintValidatorRequest) -> bool:
        return request.constraint_name == self.name

    async def check(self, request: SchemaConstraintValidatorRequest) -> list[GroupedDataPaths]:
        grouped_data_paths_list: list[GroupedDataPaths] = []

        if getattr(request.node_schema, "generate_profile", False) is True:
            return grouped_data_paths_list

        for query_class in self.query_classes:
            query = await query_class.init(
                db=self.db,
                branch=self.branch,
                node_schema=request.node_schema,
                schema_path=request.schema_path,
            )
            await query.execute(db=self.db)
            grouped_data_paths_list.append(await query.get_paths())
        return grouped_data_paths_list



from datetime import timedelta

from infrahub_sdk import InfrahubClient
from infrahub_sdk.protocols import CoreRepository
from prefect import flow, task
from prefect.automations import AutomationCore
from prefect.cache_policies import NONE
from prefect.client.orchestration import get_client
from prefect.client.schemas.filters import DeploymentFilter, DeploymentFilterName
from prefect.events.actions import RunDeployment
from prefect.events.schemas.automations import EventTrigger, Posture
from prefect.logging import get_run_logger

from infrahub import lock
from infrahub.core.constants import InfrahubKind, RepositoryInternalStatus
from infrahub.core.registry import registry
from infrahub.exceptions import RepositoryError
from infrahub.message_bus import Meta, messages
from infrahub.services import services
from infrahub.worker import WORKER_IDENTITY
from infrahub.workflows.catalogue import COMPUTED_ATTRIBUTE_SETUP_PYTHON

from ..log import get_log_data
from ..tasks.artifact import define_artifact
from ..workflows.catalogue import REQUEST_ARTIFACT_DEFINITION_GENERATE, REQUEST_ARTIFACT_GENERATE
from ..workflows.utils import add_branch_tag, add_tags
from .constants import AUTOMATION_NAME
from .models import (
    GitDiffNamesOnly,
    GitDiffNamesOnlyResponse,
    GitRepositoryAdd,
    GitRepositoryAddReadOnly,
    GitRepositoryImportObjects,
    GitRepositoryMerge,
    GitRepositoryPullReadOnly,
    RequestArtifactDefinitionGenerate,
    RequestArtifactGenerate,
)
from .repository import InfrahubReadOnlyRepository, InfrahubRepository, get_initialized_repo


@flow(
    name="git-repository-add-read-write",
    flow_run_name="Adding repository {model.repository_name} in branch {model.infrahub_branch_name}",
)
async def add_git_repository(model: GitRepositoryAdd) -> None:
    service = services.service
    await add_tags(branches=[model.infrahub_branch_name], nodes=[model.repository_id])

    async with lock.registry.get(name=model.repository_name, namespace="repository"):
        repo = await InfrahubRepository.new(
            id=model.repository_id,
            name=model.repository_name,
            location=model.location,
            client=service.client,
            infrahub_branch_name=model.infrahub_branch_name,
            internal_status=model.internal_status,
            default_branch_name=model.default_branch_name,
        )
        await repo.import_objects_from_files(
            infrahub_branch_name=model.infrahub_branch_name, git_branch_name=model.default_branch_name
        )
        if model.internal_status == RepositoryInternalStatus.ACTIVE.value:
            await repo.sync()

            # Notify other workers they need to clone the repository
            notification = messages.RefreshGitFetch(
                meta=Meta(initiator_id=WORKER_IDENTITY, request_id=get_log_data().get("request_id", "")),
                location=model.location,
                repository_id=model.repository_id,
                repository_name=model.repository_name,
                repository_kind=InfrahubKind.REPOSITORY,
                infrahub_branch_name=model.infrahub_branch_name,
                infrahub_branch_id=model.infrahub_branch_id,
            )
            await service.send(message=notification)


@flow(
    name="git-repository-add-read-only",
    flow_run_name="Adding read only repository {model.repository_name} in branch {model.infrahub_branch_name}",
)
async def add_git_repository_read_only(model: GitRepositoryAddReadOnly) -> None:
    service = services.service
    await add_tags(branches=[model.infrahub_branch_name], nodes=[model.repository_id])

    async with lock.registry.get(name=model.repository_name, namespace="repository"):
        repo = await InfrahubReadOnlyRepository.new(
            id=model.repository_id,
            name=model.repository_name,
            location=model.location,
            client=service.client,
            ref=model.ref,
            infrahub_branch_name=model.infrahub_branch_name,
        )
        await repo.import_objects_from_files(infrahub_branch_name=model.infrahub_branch_name)
        if model.internal_status == RepositoryInternalStatus.ACTIVE.value:
            await repo.sync_from_remote()

            # Notify other workers they need to clone the repository
            notification = messages.RefreshGitFetch(
                meta=Meta(initiator_id=WORKER_IDENTITY, request_id=get_log_data().get("request_id", "")),
                location=model.location,
                repository_id=model.repository_id,
                repository_name=model.repository_name,
                repository_kind=InfrahubKind.REPOSITORY,
                infrahub_branch_name=model.infrahub_branch_name,
                infrahub_branch_id=model.infrahub_branch_id,
            )
            await service.send(message=notification)


@flow(name="git-repositories-create-branch", flow_run_name="Create branch '{branch}' in Git Repositories")
async def create_branch(branch: str, branch_id: str) -> None:
    """Request to the creation of git branches in available repositories."""
    service = services.service
    await add_tags(branches=[branch])
    repositories: list[CoreRepository] = await service.client.filters(kind=CoreRepository)
    batch = await service.client.create_batch()
    for repository in repositories:
        batch.add(
            task=git_branch_create,
            client=service.client.client,
            branch=branch,
            branch_id=branch_id,
            repository_name=repository.name.value,
            repository_id=repository.id,
            repository_location=repository.location.value,
        )

    async for _, _ in batch.execute():
        pass


@flow(name="git_repositories_sync", flow_run_name="Sync Git Repositories")
async def sync_remote_repositories() -> None:
    service = services.service

    log = get_run_logger()

    branches = await service.client.branch.all()
    repositories = await service.client.get_list_repositories(branches=branches, kind=InfrahubKind.REPOSITORY)

    for repo_name, repository_data in repositories.items():
        active_internal_status = RepositoryInternalStatus.ACTIVE.value
        default_internal_status = repository_data.branch_info[registry.default_branch].internal_status
        staging_branch = None
        if default_internal_status != RepositoryInternalStatus.ACTIVE.value:
            active_internal_status = RepositoryInternalStatus.STAGING.value
            staging_branch = repository_data.get_staging_branch()

        infrahub_branch = staging_branch or registry.default_branch

        async with lock.registry.get(name=repo_name, namespace="repository"):
            init_failed = False
            try:
                repo = await InfrahubRepository.init(
                    service=service,
                    id=repository_data.repository.id,
                    name=repository_data.repository.name.value,
                    location=repository_data.repository.location.value,
                    client=service.client,
                    internal_status=active_internal_status,
                    default_branch_name=repository_data.repository.default_branch.value,
                )
            except RepositoryError as exc:
                service.log.error(str(exc))
                init_failed = True

            if init_failed:
                try:
                    repo = await InfrahubRepository.new(
                        service=service,
                        id=repository_data.repository.id,
                        name=repository_data.repository.name.value,
                        location=repository_data.repository.location.value,
                        client=service.client,
                        internal_status=active_internal_status,
                        default_branch_name=repository_data.repository.default_branch.value,
                    )
                    await repo.import_objects_from_files(
                        git_branch_name=registry.default_branch, infrahub_branch_name=infrahub_branch
                    )
                except RepositoryError as exc:
                    log.info(exc.message)
                    continue

            try:
                await repo.sync(staging_branch=staging_branch)
                # Tell workers to fetch to stay in sync
                message = messages.RefreshGitFetch(
                    meta=Meta(initiator_id=WORKER_IDENTITY, request_id=get_log_data().get("request_id", "")),
                    location=repository_data.repository.location.value,
                    repository_id=repository_data.repository.id,
                    repository_name=repository_data.repository.name.value,
                    repository_kind=repository_data.repository.get_kind(),
                    infrahub_branch_name=infrahub_branch,
                    infrahub_branch_id=branches[infrahub_branch].id,
                )
                await service.send(message=message)
            except RepositoryError as exc:
                log.info(exc.message)


@task(
    name="git-branch-create",
    task_run_name="Create branch '{branch}' in repository {repository_name}",
    cache_policy=NONE,
)
async def git_branch_create(
    client: InfrahubClient,
    branch: str,
    branch_id: str,
    repository_id: str,
    repository_name: str,
    repository_location: str,
) -> None:
    service = services.service
    log = get_run_logger()
    repo = await InfrahubRepository.init(
        id=repository_id, name=repository_name, location=repository_location, client=client
    )

    async with lock.registry.get(name=repository_name, namespace="repository"):
        await repo.create_branch_in_git(branch_name=branch, branch_id=branch_id, push_origin=True)

        # New branch has been pushed remotely, tell workers to fetch it
        message = messages.RefreshGitFetch(
            meta=Meta(initiator_id=WORKER_IDENTITY, request_id=get_log_data().get("request_id", "")),
            location=repo.get_location(),
            repository_id=str(repo.id),
            repository_name=repo.name,
            repository_kind=InfrahubKind.REPOSITORY,
            infrahub_branch_name=branch,
            infrahub_branch_id=branch_id,
        )
        await service.send(message=message)
        log.debug("Sent message to all workers to fetch the latest version of the repository (RefreshGitFetch)")


@flow(name="artifact-definition-generate", flow_run_name="Generate all artifacts")
async def generate_artifact_definition(branch: str) -> None:
    service = services.service
    await add_branch_tag(branch_name=branch)

    artifact_definitions = await service.client.all(kind=InfrahubKind.ARTIFACTDEFINITION, branch=branch, include=["id"])

    for artifact_definition in artifact_definitions:
        model = RequestArtifactDefinitionGenerate(branch=branch, artifact_definition=artifact_definition.id)
        await service.workflow.submit_workflow(
            workflow=REQUEST_ARTIFACT_DEFINITION_GENERATE, parameters={"model": model}
        )


@flow(name="artifact-generate", flow_run_name="Generate artifact {model.artifact_name}")
async def generate_artifact(model: RequestArtifactGenerate) -> None:
    service = services.service

    await add_tags(branches=[model.branch_name], nodes=[model.target_id])
    log = get_run_logger()
    repo = await get_initialized_repo(
        repository_id=model.repository_id,
        name=model.repository_name,
        service=service,
        repository_kind=model.repository_kind,
        commit=model.commit,
    )

    artifact = await define_artifact(message=model, service=service)

    try:
        result = await repo.render_artifact(artifact=artifact, message=model)
        log.debug(
            f"Generated artifact | changed: {result.changed} | {result.checksum} | {result.storage_id}",
        )
    except Exception:  # pylint: disable=broad-except
        log.exception("Failed to generate artifact")
        artifact.status.value = "Error"
        await artifact.save()


@flow(name="request_artifact_definitions_generate", flow_run_name="Trigger Generation of Artifacts for ")
async def generate_request_artifact_definition(model: RequestArtifactDefinitionGenerate) -> None:
    service = services.service
    await add_tags(branches=[model.branch])

    artifact_definition = await service.client.get(
        kind=InfrahubKind.ARTIFACTDEFINITION, id=model.artifact_definition, branch=model.branch
    )

    await artifact_definition.targets.fetch()
    group = artifact_definition.targets.peer
    await group.members.fetch()
    current_members = [member.id for member in group.members.peers]

    existing_artifacts = await service.client.filters(
        kind=InfrahubKind.ARTIFACT,
        definition__ids=[model.artifact_definition],
        include=["object"],
        branch=model.branch,
    )
    artifacts_by_member = {}
    for artifact in existing_artifacts:
        if artifact.object.id in current_members:
            artifacts_by_member[artifact.object.peer.id] = artifact.id

    await artifact_definition.transformation.fetch()
    transformation_repository = artifact_definition.transformation.peer.repository

    await transformation_repository.fetch()

    transform = artifact_definition.transformation.peer
    await transform.query.fetch()
    query = transform.query.peer
    repository = transformation_repository.peer
    branch = await service.client.branch.get(branch_name=model.branch)
    if branch.sync_with_git:
        repository = await service.client.get(
            kind=InfrahubKind.GENERICREPOSITORY, id=repository.id, branch=model.branch, fragment=True
        )
    transform_location = ""

    if transform.typename == InfrahubKind.TRANSFORMJINJA2:
        transform_location = transform.template_path.value
    elif transform.typename == InfrahubKind.TRANSFORMPYTHON:
        transform_location = f"{transform.file_path.value}::{transform.class_name.value}"

    for relationship in group.members.peers:
        member = relationship.peer
        artifact_id = artifacts_by_member.get(member.id)
        if model.limit and artifact_id not in model.limit:
            continue

        request_artifact_generate_model = RequestArtifactGenerate(
            artifact_name=artifact_definition.name.value,
            artifact_id=artifact_id,
            artifact_definition=model.artifact_definition,
            commit=repository.commit.value,
            content_type=artifact_definition.content_type.value,
            transform_type=transform.typename,
            transform_location=transform_location,
            repository_id=repository.id,
            repository_name=repository.name.value,
            repository_kind=repository.get_kind(),
            branch_name=model.branch,
            query=query.name.value,
            variables=member.extract(params=artifact_definition.parameters.value),
            target_id=member.id,
            target_name=member.display_label,
            timeout=transform.timeout.value,
        )

        await service.workflow.submit_workflow(
            workflow=REQUEST_ARTIFACT_GENERATE, parameters={"model": request_artifact_generate_model}
        )


@flow(name="git-repository-pull-read-only", flow_run_name="Pull latest commit on {model.repository_name}")
async def pull_read_only(model: GitRepositoryPullReadOnly) -> None:
    service = services.service

    await add_tags(branches=[model.infrahub_branch_name], nodes=[model.repository_id])
    log = get_run_logger()

    if not model.ref and not model.commit:
        log.warning("No commit or ref in GitRepositoryPullReadOnly message")
        return
    async with lock.registry.get(name=model.repository_name, namespace="repository"):
        init_failed = False
        try:
            repo = await InfrahubReadOnlyRepository.init(
                id=model.repository_id,
                name=model.repository_name,
                location=model.location,
                client=service.client,
                ref=model.ref,
                infrahub_branch_name=model.infrahub_branch_name,
            )
        except RepositoryError:
            init_failed = True

        if init_failed:
            repo = await InfrahubReadOnlyRepository.new(
                id=model.repository_id,
                name=model.repository_name,
                location=model.location,
                client=service.client,
                ref=model.ref,
                infrahub_branch_name=model.infrahub_branch_name,
            )

        await repo.import_objects_from_files(infrahub_branch_name=model.infrahub_branch_name, commit=model.commit)
        await repo.sync_from_remote(commit=model.commit)

        # Tell workers to fetch to stay in sync
        message = messages.RefreshGitFetch(
            meta=Meta(initiator_id=WORKER_IDENTITY, request_id=get_log_data().get("request_id", "")),
            location=model.location,
            repository_id=model.repository_id,
            repository_name=model.repository_name,
            repository_kind=InfrahubKind.READONLYREPOSITORY,
            infrahub_branch_name=model.infrahub_branch_name,
            infrahub_branch_id=model.infrahub_branch_id,
        )
        await service.send(message=message)


@flow(
    name="git-repository-merge",
    flow_run_name="Merge {model.source_branch} > {model.destination_branch} in git repository",
)
async def merge_git_repository(model: GitRepositoryMerge) -> None:
    service = services.service

    await add_tags(branches=[model.source_branch, model.destination_branch], nodes=[model.repository_id])

    repo = await InfrahubRepository.init(
        id=model.repository_id,
        name=model.repository_name,
        client=service.client,
        default_branch_name=model.default_branch,
    )

    if model.internal_status == RepositoryInternalStatus.STAGING.value:
        repo_source = await service.client.get(
            kind=InfrahubKind.GENERICREPOSITORY, id=model.repository_id, branch=model.source_branch
        )
        repo_main = await service.client.get(kind=InfrahubKind.GENERICREPOSITORY, id=model.repository_id)
        repo_main.internal_status.value = RepositoryInternalStatus.ACTIVE.value
        repo_main.sync_status.value = repo_source.sync_status.value

        commit = repo.get_commit_value(branch_name=repo.default_branch, remote=False)
        repo_main.commit.value = commit

        await repo_main.save()
    else:
        async with lock.registry.get(name=model.repository_name, namespace="repository"):
            await repo.merge(source_branch=model.source_branch, dest_branch=model.destination_branch)
            if repo.location:
                # Destination branch has changed and pushed remotely, tell workers to re-fetch
                message = messages.RefreshGitFetch(
                    meta=Meta(initiator_id=WORKER_IDENTITY, request_id=get_log_data().get("request_id", "")),
                    location=repo.location,
                    repository_id=str(repo.id),
                    repository_name=repo.name,
                    repository_kind=InfrahubKind.REPOSITORY,
                    infrahub_branch_name=model.destination_branch,
                    infrahub_branch_id=model.destination_branch_id,
                )
                await service.send(message=message)


@flow(name="git-commit-automation-setup", flow_run_name="Setup git commit updated event in task-manager")
async def setup_commit_automation() -> None:
    log = get_run_logger()

    async with get_client(sync_client=False) as client:
        deployments = {
            item.name: item
            for item in await client.read_deployments(
                deployment_filter=DeploymentFilter(
                    name=DeploymentFilterName(any_=[COMPUTED_ATTRIBUTE_SETUP_PYTHON.name])
                )
            )
        }
        if COMPUTED_ATTRIBUTE_SETUP_PYTHON.name not in deployments:
            raise ValueError("Unable to find the deployment for COMPUTED_ATTRIBUTE_SETUP_PYTHON")

        deployment_id_computed_attribute_setup_python = deployments[COMPUTED_ATTRIBUTE_SETUP_PYTHON.name].id

        schema_update_automation = await client.find_automation(id_or_name=AUTOMATION_NAME)

        automation = AutomationCore(
            name=AUTOMATION_NAME,
            description="Trigger actions on schema update event",
            enabled=True,
            trigger=EventTrigger(
                posture=Posture.Reactive,
                expect={"infrahub.repository.update_commit"},
                within=timedelta(0),
                threshold=1,
            ),
            actions=[
                RunDeployment(
                    source="selected",
                    deployment_id=deployment_id_computed_attribute_setup_python,
                    parameters={
                        "branch_name": "{{ event.resource['infrahub.branch.name'] }}",
                        "commit": "{{ event.payload['commit'] }}",
                    },
                    job_variables={},
                ),
            ],
        )

        if schema_update_automation:
            await client.update_automation(automation_id=schema_update_automation.id, automation=automation)
            log.info(f"{AUTOMATION_NAME} Updated")
        else:
            await client.create_automation(automation=automation)
            log.info(f"{AUTOMATION_NAME} Created")


@flow(name="git-repository-import-object", flow_run_name="Import objects from git repository")
async def import_objects_from_git_repository(model: GitRepositoryImportObjects) -> None:
    await add_branch_tag(model.infrahub_branch_name)
    repo = await get_initialized_repo(
        repository_id=model.repository_id,
        name=model.repository_name,
        service=services.service,
        repository_kind=model.repository_kind,
        commit=model.commit,
    )
    await repo.import_objects_from_files(infrahub_branch_name=model.infrahub_branch_name, commit=model.commit)


@flow(
    name="git-repository-diff-names-only",
    flow_run_name="Collecting modifications between commits {model.first_commit} and {model.second_commit}",
    persist_result=True,
)
async def git_repository_diff_names_only(model: GitDiffNamesOnly) -> GitDiffNamesOnlyResponse:
    service = services.service

    repo = await get_initialized_repo(
        repository_id=model.repository_id,
        name=model.repository_name,
        service=service,
        repository_kind=model.repository_kind,
    )
    files_changed: list[str] = []
    files_removed: list[str] = []

    if model.second_commit:
        files_changed, files_added, files_removed = await repo.calculate_diff_between_commits(
            first_commit=model.first_commit, second_commit=model.second_commit
        )
    else:
        files_added = await repo.list_all_files(commit=model.first_commit)

    response = GitDiffNamesOnlyResponse(
        files_added=files_added, files_changed=files_changed, files_removed=files_removed
    )
    return response



from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from prefect.events.schemas.automations import Automation  # noqa: TC002
from pydantic import BaseModel, Field
from typing_extensions import Self

if TYPE_CHECKING:
    from uuid import UUID

    from infrahub_sdk.data import RepositoryData

    from infrahub.core.schema.schema_branch_computed import PythonDefinition


class ComputedAttributeAutomations(BaseModel):
    data: dict[str, dict[str, Automation]] = Field(default_factory=lambda: defaultdict(dict))

    @classmethod
    def from_prefect(cls, automations: list[Automation], prefix: str = "") -> Self:
        obj = cls()
        for automation in automations:
            if not automation.name.startswith(prefix):
                continue

            name_split = automation.name.split("::")
            if len(name_split) != 3:
                continue

            scope = name_split[1]
            identifier = name_split[2]

            obj.data[identifier][scope] = automation

        return obj

    def get(self, identifier: str, scope: str) -> Automation:
        if identifier in self.data and scope in self.data[identifier]:
            return self.data[identifier][scope]
        raise KeyError(f"Unable to find an automation for {identifier} {scope}")

    def has(self, identifier: str, scope: str) -> bool:
        if identifier in self.data and scope in self.data[identifier]:
            return True
        return False

    def return_obsolete(self, keep: list[UUID]) -> list[UUID]:
        return [automation_id for automation_id in self.all_automation_ids if automation_id not in keep]

    @property
    def all_automation_ids(self) -> list[UUID]:
        automation_ids: list[UUID] = []
        for identifier in self.data.values():
            for automation in identifier.values():
                automation_ids.append(automation.id)
        return automation_ids


@dataclass
class PythonTransformComputedAttribute:
    name: str
    repository_id: str
    repository_name: str
    repository_kind: str
    query_name: str
    query_models: list[str]
    computed_attribute: PythonDefinition
    default_schema: bool
    branch_commit: dict[str, str] = field(default_factory=dict)

    def populate_branch_commit(self, repository_data: RepositoryData | None = None) -> None:
        if repository_data:
            for branch, commit in repository_data.branches.items():
                self.branch_commit[branch] = commit


@dataclass
class PythonTransformTarget:
    kind: str
    object_id: str



from __future__ import annotations

from collections.abc import Mapping
from typing import Any


class Error(Exception):
    def __init__(self, message: str | None = None):
        self.message = message
        super().__init__(self.message)


class JsonDecodeError(Error):
    def __init__(self, message: str | None = None, content: str | None = None, url: str | None = None):
        self.message = message
        self.content = content
        self.url = url
        if not self.message and self.url:
            self.message = f"Unable to decode response as JSON data from {self.url}"
        super().__init__(self.message)


class ServerNotReachableError(Error):
    def __init__(self, address: str, message: str | None = None):
        self.address = address
        self.message = message or f"Unable to connect to '{address}'."
        super().__init__(self.message)


class ServerNotResponsiveError(Error):
    def __init__(self, url: str, timeout: int | None = None, message: str | None = None):
        self.url = url
        self.timeout = timeout
        self.message = message or f"Unable to read from '{url}'."
        if timeout:
            self.message += f" (timeout: {timeout} sec)"
        super().__init__(self.message)


class GraphQLError(Error):
    def __init__(self, errors: list[dict[str, Any]], query: str | None = None, variables: dict | None = None):
        self.query = query
        self.variables = variables
        self.errors = errors
        self.message = f"An error occurred while executing the GraphQL Query {self.query}, {self.errors}"
        super().__init__(self.message)


class BranchNotFoundError(Error):
    def __init__(self, identifier: str, message: str | None = None):
        self.identifier = identifier
        self.message = message or f"Unable to find the branch '{identifier}' in the Database."
        super().__init__(self.message)


class SchemaNotFoundError(Error):
    def __init__(self, identifier: str, message: str | None = None):
        self.identifier = identifier
        self.message = message or f"Unable to find the schema '{identifier}'."
        super().__init__(self.message)


class ModuleImportError(Error):
    def __init__(self, message: str | None = None):
        self.message = message or "Unable to import the module"
        super().__init__(self.message)


class NodeNotFoundError(Error):
    def __init__(
        self,
        node_type: str,
        identifier: Mapping[str, list[str]],
        message: str = "Unable to find the node in the database.",
        branch_name: str | None = None,
    ):
        self.node_type = node_type
        self.identifier = identifier
        self.branch_name = branch_name

        self.message = message
        super().__init__(self.message)

    def __str__(self) -> str:
        return f"""
        {self.message}
        {self.branch_name} | {self.node_type} | {self.identifier}
        """


class ResourceNotDefinedError(Error):
    """Raised when trying to access a resource that hasn't been defined."""

    def __init__(self, message: str | None = None):
        self.message = message or "The requested resource was not found"
        super().__init__(self.message)


class InfrahubCheckNotFoundError(Error):
    def __init__(self, name: str, message: str | None = None):
        self.message = message or f"The requested InfrahubCheck '{name}' was not found."
        super().__init__(self.message)


class InfrahubTransformNotFoundError(Error):
    def __init__(self, name: str, message: str | None = None):
        self.message = message or f"The requested InfrahubTransform '{name}' was not found."
        super().__init__(self.message)


class ValidationError(Error):
    def __init__(self, identifier: str, message: str):
        self.identifier = identifier
        self.message = message
        super().__init__(self.message)


class AuthenticationError(Error):
    def __init__(self, message: str | None = None):
        self.message = message or "Authentication Error, unable to execute the query."
        super().__init__(self.message)


class FeatureNotSupportedError(Error):
    """Raised when trying to use a method on a node that doesn't support it."""


class UninitializedError(Error):
    """Raised when an object requires an initialization step before use"""


class InvalidResponseError(Error):
    """Raised when an object requires an initialization step before use"""
